from math import *

import pygame as pg

import datetime, os
import os.path
import subprocess

import numpy as np

import bluesky as bs
from bluesky.core import Entity
from bluesky import stack
from bluesky.tools import geo, areafilter
from bluesky.tools.aero import ft, kts, nm
from bluesky.tools.misc import tim2txt
from . import splash
from .keyboard import Keyboard
from .fastfont import Fastfont
from .console import Console
from .menu import Menu
from .dialog import fileopen as opendialog

black    = (0, 0, 0)
white    = (255,255,255)
darkgrey = (25, 25, 48)
grey     = (84, 84, 114)
darkblue = (25, 25, 64, 100)
white    = (255, 255, 255)
green    = (0, 255, 0)
blue     = (0, 0, 255)
red      = (255, 0, 0)
cyan     = (0,150,150)
lightgreyblue = (130, 150, 190)  # waypoint symbol color
lightgreygreen = (149, 215, 179)  # grid color
lightcyan = (0, 255, 255)  # FIR boundaries
amber    = (255,163,71)  # Conflicting aircraft
magenta  = (255,0,255) # Used for route

class Screen(Entity):
    """
    Screen class definition : contains radar & edit screen data & methods

    Methods:
        Screen(tmx)         :  constructor

        echo(msg)           : print something at screen
        update()            : Draw a new frame of screen
        ll2xy(lat,lon)      : lat/lon[deg] to pixel coordinate conversion
        xy2ll(x,y)          : pixel to lat/lon[de]g conversion
        zoom(factor)        : zoom in/out
        pan(lat,lon,txt)    : pan to lat,lon position

    Members: see constructor

    Created by : Jacco M. Hoekstra (TU Delft)
    Updated by : Jerom Maas

    """
    def __init__(self):
        super().__init__()
        # processes input from keyboard & mouse
        self.keyb = Keyboard()

        # Parameters for making screenshots
        self.session = "new"
        self.folder  = ""
        self.screenshot = False
        self.screenshotname = ""

        # Isometric display parameter
        self.isoalt = 0.  # how many meters one pixel is high

        # Display ADS-B range flag
        self.swAdsbCoverage = False

        # Update rate radar:
        self.radardt = 0.10  # 10x per sec 0.25  # 4x per second max
        self.radt0 = -999.  # last time drawn
        self.maxnrwp = 1000  # max nr apts+wpts to be drawn

        # Keep track of sim dt to show average update rate on screen
        self.dts = []

    def reset(self):
        self.objdel()     # Delete user defined objects

    def init(self):
        # Read Screen configuration file:
        print()
        print("Setting up screen...")

        lst = np.genfromtxt(bs.resource("graphics/scr_cfg.dat"), comments='#', dtype='i4')

        self.swfullscreen = int(lst[0]) == 0

        self.width = int(lst[1])  # default value to create variable
        self.height = int(lst[2])  # default value to create variable

        # Dimensions radar window
        self.lat1 = 53.  # [deg] upper limit display
        self.lat0 = 51.  # [deg] lowerlimit display
        self.lon0 = -1.  # [deg] left side of screen

        dellat = self.lat1 - self.lat0

        avelat = (self.lat0 + self.lat1) * 0.5

        dellon = dellat * self.width / (self.height * cos(radians(avelat)))
        avelon = (self.lon0 + dellon / 2. + 180.) % 360. - 180.

        self.lon1 = (self.lon0 + dellon + 180.) % 360. - 180.

        self.ctrlat = avelat
        self.ctrlon = avelon


        #----------------------------SYMBOLS-----------------------------
        # Read graphics for acsymbol (normal = green) + amber
        self.acsymbol = []
        for i in range(60):
            self.acsymbol.append(pg.image.load(bs.resource("graphics/acsymbol/acsymbol" \
                                  + str(i) + ".png")))

        self.ambacsymbol = []
        for i in range(60):
            self.ambacsymbol.append(pg.image.load(bs.resource("graphics/amb-acsymbol/amb-acsymbol" \
                                     + str(i) + ".png")))

        # Lable lines& default no trails
        self.swlabel = 3


        # Read and scale waypoint symbol
        wptgif = pg.image.load(bs.resource("graphics/waypoint.gif"))
        self.wptsymbol = pg.transform.scale(wptgif, (10, 7))
        self.wpsw = 1  # 0=None, 1 = VOR 2 = non-digit ones, 3 =all

        # Read and scale airport symbol
        aptgif = pg.image.load(bs.resource("graphics/airport.gif"))
        self.aptsymbol = pg.transform.scale(aptgif, (12, 9))
        self.apsw = 1  # 0 = None, 1 = Large, 2 = All

        # Free flight displays
        self.swsep = False # To show circles of 2.5 nm radius around each aircraft
                           # Note: circles will be distorted when away from equator
        self.swspd = False # To show speed vectors of each aircraft
        self.swtestarea= False

        #--------------------------------MAPS---------------------------------
        # Read map of world
        self.mapbmp = pg.image.load(bs.resource("graphics/world.jpg"))
        w, h = self.mapbmp.get_size()

        # Two ref positions for scaling, convert to scaling factors x=a*lat+b
        x1, y1, lat1, lon1 = 0., 0., 90., -180.
        x2, y2, lat2, lon2 = w, h, -90., 180.
        self.mapax = (x2 - x1) / (lon2 - lon1)
        self.mapbx = x2 - self.mapax * lon2
        self.mapay = (y2 - y1) / (lat2 - lat1)
        self.mapby = y2 - self.mapay * lat2

        self.swsat = True

        # Nav display projection mode
        self.swnavdisp = False
        self.ndacid = ""
        self.ndlat = 0.0
        self.ndlon = 0.0
        self.ndhdg = 0.0

        #------------------------WINDOW SETUP and scaling--------------------------
        # Empty tuple to force reselection waypoints & airports to be drawn
        self.navsel = ()  # Empty tuple to force reselection waypoints to be drawn
        self.satsel = ()  # Empty tuple to force reselection satellite imagery to be drawn


        # Set up window
        splash.destroy()  # does pg.display.quit()!
        pg.display.init()

        # Full screen
        di = pg.display.Info()
        if self.swfullscreen:
            self.width = di.current_w
            self.height = di.current_h
            reso = (self.width, self.height)
            self.win = pg.display.set_mode(reso, pg.FULLSCREEN)
        else:
            # Windowed
            self.height = int(min(self.height, int(di.current_h * 90 / 100)))
            self.width = int(min(self.width, int(di.current_w * 90 / 100)))
            reso = (self.width, self.height)
            self.win = pg.display.set_mode(reso)

        pg.display.set_caption("BlueSky Open ATM Simulator (F11 = Full Screen)", "BlueSky")
        iconpath = bs.resource("graphics/icon.gif")
        iconbmp = pg.image.load(iconpath)
        pg.display.set_icon(iconbmp)

        self.radbmp = self.win.copy()
        self.redrawradbg = True  # Switch to redraw background

        #---------------------RADAR FONTS & EDIT WINDOW-----------------------------
        # Set up fonts
        self.fontrad = Fastfont(self.win, 'Arial', 14, green, False, False)  # name, size, bold,italic
        self.fontamb = Fastfont(self.win, 'Arial', 14, amber, False, False)  # name, size, bold,italic
        self.fontnav = Fastfont(self.win, 'Arial', 12, lightgreyblue, False, False)  # name, size, bold,italic
        self.fontsys = Fastfont(self.win, 'Helvetica', 14, white, False, False)  # name, size, bold,italic

        # Edit window: 6 line of 64 chars
        nch = lst[3]  # number of chars per line
        nlin = lst[4]  # number of lines in windows
        winx = lst[5]  # x-coordinate in pixels of left side
        winy = self.height - lst[6]  # y-coordinate in pixels of bottom
        self.editwin = Console(self.win, nch, nlin, winx, winy)

        # Menu button window
        self.menu = Menu(self.win,10,36)


        #-------------------------COASTLINE DATA--------------------------------------
        # Init geo (coastline)  data
        f = open(bs.resource("navdata/coastlines.dat"), 'r')
        print("Reading coastlines.dat")
        lines = f.readlines()
        f.close()
        records = []
        for line in lines:
            if not (line.strip() == "" or line.strip()[0] == '#'):
                arg = line.split()
                if len(arg) == 3:
                    records.append([arg[0], float(arg[1]), float(arg[2])])

        # print len(records), " records read."
        # Convert pen up/pen down format of coastlines to numpy arrays with lat/lon

        coastlat0 = []
        coastlon0 = []
        coastlat1 = []
        coastlon1 = []
        clat, clon = -1, -1

        for rec in records:
            lat, lon = rec[1], rec[2]
            if rec[0] == 'D':
                coastlat0.append(clat)
                coastlon0.append(clon)
                coastlat1.append(lat)
                coastlon1.append(lon)
            clat, clon = lat, lon

        self.coastlat0 = np.array(coastlat0)
        self.coastlon0 = np.array(coastlon0)
        self.coastlat1 = np.array(coastlat1)
        self.coastlon1 = np.array(coastlon1)

        del coastlat0, coastlon0, coastlat1, coastlon1  # Clear memory

        self.geosel = ()  # Force reselect first time coastlines
        self.firsel = ()  # Force reselect first time FIR lines

        print("    ", len(self.coastlat0), " coastlines added.")

        # Set default coastlines & FIRs on:
        self.swgeo = True
        self.swfir = True
        self.swgrid = False

        # Route drawing for which acid? "" =  None
        self.acidrte = ""
        self.rtewpid = []
        self.rtewplabel = []

        # User defined background objects
        self.objtype    = []
        self.objcolor   = []
        self.objdata    = []
        self.objname    = []

        # Wpt and apt drawing logic memory
        self.wpswbmp = []              # switch indicating whether label bmp is present
        self.wplabel = []              # List to store bitmaps of label
        self.apswbmp = []              # switch indicating whether label bmp is present
        self.aplabel = []              # List to store bitmaps of label

        self.updateNavBuffers()
        return

    def shownd(self, acid):
        if acid:
            self.ndacid = acid
        self.swnavdisp = not self.swnavdisp

    def updateNavBuffers(self):
        self.wpswbmp = len(bs.navdb.wplat) * [False]
        self.wplabel = len(bs.navdb.wplat) * [0]

        self.apswbmp = len(bs.navdb.aptlat) * [False]
        self.aplabel = len(bs.navdb.aptlat) * [0]

    @stack.command(annotations='string')
    def echo(self, msg='', flags=0):
        ''' Echo a message to the BlueSky console window. '''
        if msg:
            msgs = msg.split('\n')
            for m in msgs:
                self.editwin.echo(m)
        return

    @stack.command(name='INSEDIT')
    def cmdline(self, text: 'string'):
        ''' Insert text op edit line in command window. '''
        self.editwin.insert(text)

    def update(self):
        """Draw a new frame"""
        # First check for keys & mouse
        self.keyb.update()
        # Navdisp mode: get center:
        if self.swnavdisp:
            i = bs.traf.id2idx(self.ndacid)
            if i >= 0:
                self.ndlat = bs.traf.lat[i]
                self.ndlon = bs.traf.lon[i]
                self.ndcrs = bs.traf.hdg[i]
            else:
                self.swnavdisp = False
        else:
            self.ndcrs = 0.0

        # Simulation: keep track of timestep
        # For measuring game loop frequency
        self.dts.append(bs.sim.simdt)
        if len(self.dts) > 20:
                del self.dts[0]

        # Radar window
        # --------------Background--------------

        if self.redrawradbg or self.swnavdisp:
            if self.swnavdisp or not self.swsat:
                self.radbmp.fill(darkgrey)

            else:
                #--------------Satellite image--------------
                navsel = (self.lat0, self.lat1, \
                          self.lon0, self.lon1)
                if not self.satsel == navsel:
                    # Map cutting and scaling: normal case
                    if self.lon1 > self.lon0:
                        x0 = max(0, self.lon0 * self.mapax + self.mapbx)
                        x1 = min(self.mapbmp.get_width() - 1, \
                                 self.lon1 * self.mapax + self.mapbx)

                        y1 = min(self.mapbmp.get_height() - 1, \
                                 self.lat0 * self.mapay + self.mapby)
                        y0 = max(0, self.lat1 * self.mapay + self.mapby)

                        selrect = pg.Rect(x0, y0, abs(x1 - x0), abs(y1 - y0))
                        mapsel = self.mapbmp.subsurface(selrect)
                        self.submap = pg.transform.scale(mapsel, \
                                                         (self.width, self.height))

                        self.radbmp.blit(self.submap, (0, 0))

                    else:
                        # Wrap around case: clip two segments
                        w0 = int(self.width * (180. - self.lon0) / \
                                 (180.0 - self.lon0 + self.lon1 + 180.))
                        w1 = int(self.width - w0)

                        # Left part
                        x0 = max(0, self.lon0 * self.mapax + self.mapbx)
                        x1 = self.mapbmp.get_width() - 1

                        y1 = min(self.mapbmp.get_height() - 1, \
                                 self.lat0 * self.mapay + self.mapby)
                        y0 = max(0, self.lat1 * self.mapay + self.mapby)

                        selrect = pg.Rect(x0, y0, abs(x1 - x0), abs(y1 - y0))
                        mapsel = self.mapbmp.subsurface(selrect)
                        self.submap = pg.transform.scale(mapsel, \
                                                         (w0, self.height))
                        self.radbmp.blit(self.submap, (0, 0))

                        # Right half
                        x0 = 0
                        x1 = min(self.mapbmp.get_width() - 1, \
                                 self.lon1 * self.mapax + self.mapbx)

                        selrect = pg.Rect(x0, y0, abs(x1 - x0), abs(y1 - y0))
                        mapsel = self.mapbmp.subsurface(selrect)
                        self.submap = pg.transform.scale(mapsel, \
                                                         (w1, self.height))
                        self.radbmp.blit(self.submap, (w0, 0))
                        self.submap = self.radbmp.copy()

                    self.satsel = navsel

                else:
                    # Map blit only
                    self.radbmp.blit(self.submap, (0, 0))


            if self.swgrid and not self.swnavdisp:
                # ------Draw lat/lon grid------
                ngrad = int(self.lon1 - self.lon0)

                if ngrad >= 10:
                    step = 10
                    i0 = step * int(self.lon0 / step)
                    j0 = step * int(self.lat0 / step)
                else:
                    step = 1
                    i0 = int(self.lon0)
                    j0 = int(self.lon0)

                for i in range(i0, int(self.lon1 + 1.), step):
                    x, y = self.ll2xy(self.ctrlat, i)
                    pg.draw.line(self.radbmp, lightgreygreen, \
                                 (x, 0), (x, self.height))

                for j in range(j0, int(self.lat1 + 1.), step):
                    x, y = self.ll2xy(j, self.ctrlon)
                    pg.draw.line(self.radbmp, lightgreygreen, \
                                 (0, y), (self.width, y))

            #------ Draw coastlines ------
            if self.swgeo:
                # cx,cy = -1,-1
                geosel = (self.lat0, self.lon0, self.lat1, self.lon1)
                if self.geosel != geosel:
                    self.geosel = geosel

                    self.cstsel = np.where(
                        self.onradar(self.coastlat0, self.coastlon0) + \
                        self.onradar(self.coastlat1, self.coastlon1))

                    # print len(self.cstsel[0])," coastlines"
                    self.cx0, self.cy0 = self.ll2xy(self.coastlat0, self.coastlon0)
                    self.cx1, self.cy1 = self.ll2xy(self.coastlat1, self.coastlon1)

                for i in list(self.cstsel[0]):
                    pg.draw.line(self.radbmp, grey, (self.cx0[i], self.cy0[i]), \
                                 (self.cx1[i], self.cy1[i]))

            #------ Draw FIRs ------
            if self.swfir:
                self.firx0, self.firy0 = self.ll2xy(bs.navdb.firlat0, \
                                                    bs.navdb.firlon0)

                self.firx1, self.firy1 = self.ll2xy(bs.navdb.firlat1, \
                                                    bs.navdb.firlon1)

                for i in range(len(self.firx0)):
                    pg.draw.line(self.radbmp, lightcyan,
                                 (self.firx0[i], self.firy0[i]),
                                 (self.firx1[i], self.firy1[i]))

            # -----------------Waypoint & airport symbols-----------------
            # Check whether we need to reselect waypoint set to be drawn

            navsel = (self.lat0, self.lat1, \
                      self.lon0, self.lon1)
            if self.navsel != navsel:
                self.navsel = navsel

                # Make list of indices of waypoints & airports on screen

                self.wpinside = list(np.where(self.onradar(bs.navdb.wplat, \
                                                           bs.navdb.wplon))[0])

                self.wptsel = []
                for i in self.wpinside:
                    if self.wpsw == 3 or \
                            (self.wpsw == 1 and len(bs.navdb.wpid[i]) == 3) or \
                            (self.wpsw == 2 and bs.navdb.wpid[i].isalpha()):
                        self.wptsel.append(i)
                self.wptx, self.wpty = self.ll2xy(bs.navdb.wplat, bs.navdb.wplon)

                self.apinside = list(np.where(self.onradar(bs.navdb.aptlat, \
                                                           bs.navdb.aptlon))[0])

                self.aptsel = []
                for i in self.apinside:
                    if self.apsw == 2 or (self.apsw == 1 and \
                                                      bs.navdb.aptmaxrwy[i] > 1000.):
                        self.aptsel.append(i)
                self.aptx, self.apty = self.ll2xy(bs.navdb.aptlat, bs.navdb.aptlon)


            #------- Draw waypoints -------
            if self.wpsw > 0:
                # print len(self.wptsel)," waypoints"
                if len(self.wptsel) < self.maxnrwp:
                    wptrect = self.wptsymbol.get_rect()
                    for i in self.wptsel:
                        # wptrect.center = self.ll2xy(bs.navdb.wplat[i],  \
                        #     bs.navdb.wplon[i])
                        wptrect.center = self.wptx[i], self.wpty[i]
                        self.radbmp.blit(self.wptsymbol, wptrect)

                        # If waypoint label bitmap does not yet exist, make it
                        if not self.wpswbmp[i]:
                            self.wplabel[i] = pg.Surface((80, 30), 0, self.win)
                            self.fontnav.printat(self.wplabel[i], 0, 0, \
                                                 bs.navdb.wpid[i])
                            self.wpswbmp[i] = True

                        # In any case, blit it
                        xtxt = wptrect.right + 2
                        ytxt = wptrect.top
                        self.radbmp.blit(self.wplabel[i], (xtxt, ytxt), \
                                         None, pg.BLEND_ADD)

                        if not self.wpswbmp[i]:
                            xtxt = wptrect.right + 2
                            ytxt = wptrect.top

                            # self.fontnav.printat(self.radbmp,xtxt,ytxt, \
                            #     bs.navdb.wpid[i])

            #------- Draw airports -------
            if self.apsw > 0:
                # if len(self.aptsel)<800:
                aptrect = self.aptsymbol.get_rect()

                # print len(self.aptsel)," airports"

                for i in self.aptsel:
                    # aptrect.center = self.ll2xy(bs.navdb.aptlat[i],  \
                    #                            bs.navdb.aptlon[i])
                    aptrect.center = self.aptx[i], self.apty[i]
                    self.radbmp.blit(self.aptsymbol, aptrect)

                    # If airport label bitmap does not yet exist, make it
                    if not self.apswbmp[i]:
                        self.aplabel[i] = pg.Surface((50, 30), 0, self.win)
                        self.fontnav.printat(self.aplabel[i], 0, 0, \
                                             bs.navdb.aptid[i])
                        self.apswbmp[i] = True

                    # In either case, blit it
                    xtxt = aptrect.right + 2
                    ytxt = aptrect.top
                    self.radbmp.blit(self.aplabel[i], (xtxt, ytxt), \
                                     None, pg.BLEND_ADD)

                    # self.fontnav.printat(self.radbmp,xtxt,ytxt, \
                    #     bs.navdb.aptid[i])


            #---------- Draw background trails ----------
            if bs.traf.trails.active:
                bs.traf.trails.buffer()  # move all new trails to background

                trlsel = list(np.where(
                    self.onradar(bs.traf.trails.bglat0, bs.traf.trails.bglon0) + \
                    self.onradar(bs.traf.trails.bglat1, bs.traf.trails.bglon1))[0])

                x0, y0 = self.ll2xy(bs.traf.trails.bglat0, bs.traf.trails.bglon0)
                x1, y1 = self.ll2xy(bs.traf.trails.bglat1, bs.traf.trails.bglon1)

                for i in trlsel:
                    pg.draw.aaline(self.radbmp, bs.traf.trails.bgcol[i], \
                                   (x0[i], y0[i]), (x1[i], y1[i]))

            #---------- Draw ADSB Coverage Area
            if self.swAdsbCoverage:
                # These points are based on the positions of the antennas with range = 200km
                adsbCoverageLat = [53.7863,53.5362,52.8604,51.9538,51.2285,50.8249,50.7382,
                                   50.9701,51.6096,52.498,53.4047,53.6402]
                adsbCoverageLon = [4.3757,5.8869,6.9529,7.2913,6.9312,6.251,5.7218,4.2955,
                                   3.2162,2.7701,3.1117,3.4891]

                for i in range(0,len(adsbCoverageLat)):
                    if i == len(adsbCoverageLat)-1:
                        x0, y0 = self.ll2xy(adsbCoverageLat[i],adsbCoverageLon[i])
                        x1, y1 = self.ll2xy(adsbCoverageLat[0],adsbCoverageLon[0])

                    else:
                        x0, y0 = self.ll2xy(adsbCoverageLat[i],adsbCoverageLon[i])
                        x1, y1 = self.ll2xy(adsbCoverageLat[i+1],adsbCoverageLon[i+1])

                    pg.draw.line(self.radbmp, red,(x0, y0), (x1, y1))

            # User defined objects
            for i in range(len(self.objtype)):

                # Draw LINE or POLYGON with objdata = [lat0,lon,lat1,lon1,lat2,lon2,..]
                if self.objtype[i]=='LINE' or self.objtype[i]=="POLY" or self.objtype[i]=="POLYLINE":
                    npoints = int(len(self.objdata[i])/2)
                    print(npoints)
                    x0,y0 = self.ll2xy(self.objdata[i][0],self.objdata[i][1])
                    for j in range(1,npoints):
                        x1,y1 = self.ll2xy(self.objdata[i][j*2],self.objdata[i][j*2+1])
                        pg.draw.line(self.radbmp,self.objcolor[i],(x0, y0), (x1, y1))
                        x0,y0 = x1,y1

                    if self.objtype[i]=="POLY":
                        x1,y1 = self.ll2xy(self.objdata[i][0],self.objdata[i][1])
                        pg.draw.line(self.radbmp,self.objcolor[i],(x0, y0), (x1, y1))

                # Draw bounding box of objdata = [lat0,lon0,lat1,lon1]
                elif self.objtype[i]=='BOX':
                    lat0 = min(self.objdata[i][0],self.objdata[i][2])
                    lon0 = min(self.objdata[i][1],self.objdata[i][3])
                    lat1 = max(self.objdata[i][0],self.objdata[i][2])
                    lon1 = max(self.objdata[i][1],self.objdata[i][3])

                    x0,y0 = self.ll2xy(lat1,lon0)
                    x1,y1 = self.ll2xy(lat0,lon1)
                    pg.draw.rect(self.radbmp,self.objcolor[i],pg.Rect(x0, y0, x1-x0, y1-y0),1)

                # Draw circle with objdata = [latcenter,loncenter,radiusnm]
                elif self.objtype[i]=='CIRCLE':
                    pass
                    xm,ym     = self.ll2xy(self.objdata[i][0],self.objdata[i][1])
                    xtop,ytop = self.ll2xy(self.objdata[i][0]+self.objdata[i][2]/60.,self.objdata[i][1])
                    radius    = int(round(abs(ytop-ym)))
                    pg.draw.circle(self.radbmp, self.objcolor[i], (int(xm),int(ym)), radius, 1)

            # Reset background drawing switch
            self.redrawradbg = False

        ##############################################################################
        #                          END OF BACKGROUND DRAWING                         #
        ##############################################################################

        # Blit background
        self.win.blit(self.radbmp, (0, 0))

        # Decide to redraw radar picture of a/c
        syst = pg.time.get_ticks() * 0.001
        redrawrad = self.redrawradbg or abs(syst - self.radt0) >= self.radardt

        if redrawrad:
            self.radt0 = syst  # Update lats drawing time of radar


            # Select which aircraft are within screen area
            trafsel = np.where((bs.traf.lat > self.lat0) * (bs.traf.lat < self.lat1) * \
                               (bs.traf.lon > self.lon0) * (bs.traf.lon < self.lon1))[0]

            # ------------------- Draw aircraft -------------------
            # Convert lat,lon to x,y

            trafx, trafy = self.ll2xy(bs.traf.lat, bs.traf.lon)
            trafy -= bs.traf.alt*self.isoalt

            if bs.traf.trails.active:
                ltx, lty = self.ll2xy(bs.traf.trails.lastlat, bs.traf.trails.lastlon)

            # Find pixel size of horizontal separation on screen
            # pixelrad=self.dtopix_eq(bs.traf.cd.rpz/2)
            pixelrad = self.dtopix_eq(5 * nm / 2)

            # Loop through all traffic indices which we found on screen
            for i in trafsel:

                # Get index of ac symbol, based on heading and its rect object
                isymb = int(round((bs.traf.hdg[i] - self.ndcrs) / 6.)) % 60
                pos = self.acsymbol[isymb].get_rect()

                # Draw aircraft symbol
                pos.centerx = trafx[i]
                pos.centery = trafy[i]
                dy = int(self.fontrad.linedy * 7 / 6)

                # Draw aircraft altitude line
                if self.isoalt>1e-7:
                    pg.draw.line(self.win,white,(int(trafx[i]),int(trafy[i])),(int(trafx[i]),int(trafy[i]+bs.traf.alt[i]*self.isoalt)))

                # Normal symbol if no conflict else amber
                toosmall=self.lat1-self.lat0>6 #don't draw circles if zoomed out too much

                if not bs.traf.cd.inconf[i]:
                    self.win.blit(self.acsymbol[isymb], pos)
                    if self.swsep and not toosmall:
                        pg.draw.circle(self.win,green,(int(trafx[i]),int(trafy[i])),pixelrad,1)
                else:
                    self.win.blit(self.ambacsymbol[isymb], pos)
                    if self.swsep and not toosmall:
                        pg.draw.circle(self.win,amber,(int(trafx[i]),int(trafy[i])),pixelrad,1)


                # Draw last trail part
                if bs.traf.trails.active:
                    pg.draw.line(self.win, tuple(bs.traf.trails.accolor[i]),
                                 (ltx[i], lty[i]), (trafx[i], trafy[i]))

                # Label text
                label = []
                if self.swlabel > 0:
                    label.append(bs.traf.id[i])  # Line 1 of label: id
                else:
                    label.append(" ")
                if self.swlabel > 1:
                    if bs.traf.alt[i]>bs.traf.translvl:
                        label.append("FL"+str(int(round(bs.traf.alt[i] / (100.*ft)))))  # Line 2 of label: altitude
                    else:
                        label.append(str(int(round(bs.traf.alt[i] / ft))))  # Line 2 of label: altitude
                else:
                    label.append(" ")
                if self.swlabel > 2:
                    cas = bs.traf.cas[i] / kts
                    label.append(str(int(round(cas))))  # line 3 of label: speed
                else:
                    label.append(" ")


                # Check for changes in traffic label text
                if  not (type(bs.traf.label[i])==list) or \
                      not (type(bs.traf.label[i][3])==str) or \
                        not (label[:3] == bs.traf.label[i][:3]):

                    bs.traf.label[i] = []
                    labelbmp = pg.Surface((100, 60), 0, self.win)
                    if not bs.traf.cd.inconf[i]:
                        acfont = self.fontrad
                    else:
                        acfont = self.fontamb

                    acfont.printat(labelbmp, 0, 0, label[0])
                    acfont.printat(labelbmp, 0, dy, label[1])
                    acfont.printat(labelbmp, 0, 2 * dy, label[2])

                    bs.traf.label[i].append(label[0])
                    bs.traf.label[i].append(label[1])
                    bs.traf.label[i].append(label[2])
                    bs.traf.label[i].append(labelbmp)

                # Blit label
                dest = bs.traf.label[i][3].get_rect()
                dest.top = trafy[i] - 5
                dest.left = trafx[i] + 15
                self.win.blit(bs.traf.label[i][3], dest, None, pg.BLEND_ADD)

                # Draw aircraft speed vectors
                if self.swspd:
                    # just a nominal length: a speed of 150 kts will be displayed
                    # as an arrow of 30 pixels long on screen
                    nomlength    = 30
                    nomspeed     = 150.

                    vectorlength = float(nomlength)*bs.traf.tas[i]/nomspeed

                    spdvcx = trafx[i] + np.sin(np.radians(bs.traf.trk[i])) * vectorlength
                    spdvcy = trafy[i] - np.cos(np.radians(bs.traf.trk[i])) * vectorlength \
                                - bs.traf.vs[i]/nomspeed*nomlength*self.isoalt /   \
                                            self.dtopix_eq(1e5)*1e5

                    pg.draw.line(self.win,green,(trafx[i],trafy[i]),(spdvcx,spdvcy))

            # ---- End of per aircraft i loop


            # Draw conflicts: line from a/c to closest point of approach
            nconf = len(bs.traf.cd.confpairs_unique)
            n2conf = len(bs.traf.cd.confpairs)

            if nconf>0:

                for j in range(n2conf):
                    i = bs.traf.id2idx(bs.traf.cd.confpairs[j][0])
                    if i>=0 and i<bs.traf.ntraf and (i in trafsel):
                        latcpa, loncpa = geo.kwikpos(bs.traf.lat[i], bs.traf.lon[i], \
                                                    bs.traf.trk[i], bs.traf.cd.tcpamax[j] * bs.traf.gs[i] / nm)
                        altcpa = bs.traf.lat[i] + bs.traf.vs[i]*bs.traf.cd.tcpamax[j]
                        xc, yc = self.ll2xy(latcpa,loncpa)
                        yc = yc - altcpa * self.isoalt
                        pg.draw.line(self.win,amber,(xc,yc),(trafx[i],trafy[i]))

            # Draw selected route:
            if self.acidrte != "":
                i = bs.traf.id2idx(self.acidrte)
                if i >= 0:
                    for j in range(0,bs.traf.ap.route[i].nwp):
                        if j==0:
                            x1,y1 = self.ll2xy(bs.traf.ap.route[i].wplat[j], \
                                               bs.traf.ap.route[i].wplon[j])
                        else:
                            x0,y0 = x1,y1
                            x1,y1 = self.ll2xy(bs.traf.ap.route[i].wplat[j], \
                                               bs.traf.ap.route[i].wplon[j])
                            pg.draw.line(self.win, magenta,(x0,y0),(x1,y1))

                        if j>=len(self.rtewpid) or not self.rtewpid[j]== bs.traf.ap.route[i].wpname[j]:
                            # Waypoint name labels
                            # If waypoint label bitmap does not yet exist, make it

                            # Waypoint name and constraint(s), if there are any
                            txt = bs.traf.ap.route[i].wpname[j]

                            alt = bs.traf.ap.route[i].wpalt[j]
                            spd = bs.traf.ap.route[i].wpspd[j]

                            if alt>=0. or spd >=0.:
                                # Altitude
                                if alt < 0:
                                    txt = txt + " -----/"

                                elif alt > 4500 * ft:
                                    FL = int(round((alt/(100.*ft))))
                                    txt = txt+" FL"+str(FL)+"/"

                                else:
                                    txt = txt+" "+str(int(round(alt / ft))) + "/"

                                # Speed
                                if spd < 0:
                                    txt = txt+"---"
                                else:
                                    txt = txt+str(int(round(spd / kts)))

                            wplabel = pg.Surface((128, 32), 0, self.win)
                            self.fontnav.printat(wplabel, 0, 0, \
                                                 txt)

                            if j>=len(self.rtewpid):
                                self.rtewpid.append(txt)
                                self.rtewplabel.append(wplabel)
                            else:
                                self.rtewpid[j]    = txt
                                self.rtewplabel[j] = wplabel

                        # In any case, blit the waypoint name
                        xtxt = x1 + 7
                        ytxt = y1 - 3
                        self.radbmp.blit(self.rtewplabel[j], (xtxt, ytxt), \
                                         None, pg.BLEND_ADD)

                        # Line from aircraft to active waypoint
                        if bs.traf.ap.route[i].iactwp == j:
                            x0,y0 = self.ll2xy(bs.traf.lat[i],bs.traf.lon[i])
                            pg.draw.line(self.win, magenta,(x0,y0),(x1,y1))



            # Draw aircraft trails which are on screen
            if bs.traf.trails.active:
                trlsel = list(np.where(
                    self.onradar(bs.traf.trails.lat0, bs.traf.trails.lon0) + \
                    self.onradar(bs.traf.trails.lat1, bs.traf.trails.lon1))[0])

                x0, y0 = self.ll2xy(bs.traf.trails.lat0, bs.traf.trails.lon0)
                x1, y1 = self.ll2xy(bs.traf.trails.lat1, bs.traf.trails.lon1)

                for i in trlsel:
                    pg.draw.line(self.win, bs.traf.trails.col[i], \
                                 (x0[i], y0[i]), (x1[i], y1[i]))

                # Redraw background => buffer ; if >1500 foreground linepieces on screen
                if len(trlsel) > 1500:
                    self.redrawradbg = True

        # Draw edit window
        self.editwin.update()

        if self.redrawradbg or redrawrad or self.editwin.redraw:
            self.win.blit(self.menu.bmps[self.menu.ipage], \
                           (self.menu.x, self.menu.y))
            self.win.blit(self.editwin.bmp, (self.editwin.winx, self.editwin.winy))

            # Draw frames
            pg.draw.rect(self.win, white, self.editwin.rect, 1)
            pg.draw.rect(self.win, white, pg.Rect(1, 1, self.width - 1, self.height - 1), 1)

            # Add debug line
            self.fontsys.printat(self.win, 10, 2, str(bs.sim.utc.replace(microsecond=0)))
            self.fontsys.printat(self.win, 10, 18, tim2txt(bs.sim.simt))
            self.fontsys.printat(self.win, 10+80, 2, \
                                 "ntraf = " + str(bs.traf.ntraf))
            self.fontsys.printat(self.win, 10+160, 2, \
                                 "Freq=" + str(int(len(self.dts) / max(0.001, sum(self.dts)))))

            self.fontsys.printat(self.win, 10+240, 2, \
                                 "#LOS      = " + str(len(bs.traf.cd.lospairs_unique)))
            self.fontsys.printat(self.win, 10+240, 18, \
                                 "Total LOS = " + str(len(bs.traf.cd.lospairs_all)))
            self.fontsys.printat(self.win, 10+240, 34, \
                                 "#Con      = " + str(len(bs.traf.cd.confpairs_unique)))
            self.fontsys.printat(self.win, 10+240, 50, \
                                 "Total Con = " + str(len(bs.traf.cd.confpairs_all)))

            # Frame ready, flip to screen
            pg.display.flip()

            # If needed, take a screenshot
            if self.screenshot:
                pg.image.save(self.win,self.screenshotname)
                self.screenshot=False
        return


    def ll2xy(self, lat, lon):
        if not self.swnavdisp:
            # RADAR mode:
            # Convert lat/lon to pixel x,y

            # Normal case
            if self.lon1 > self.lon0:
                x = self.width * (lon - self.lon0) / (self.lon1 - self.lon0)

            # Wrap around:
            else:
                dellon = 180. - self.lon0 + self.lon1 + 180.
                xlon = lon + (lon < 0.) * 360.
                x = (xlon - self.lon0) / dellon * self.width

            y = self.height * (self.lat1 - lat) / (self.lat1 - self.lat0)
        else:
            # NAVDISP mode:
            qdr, dist = geo.qdrdist(self.ndlat, self.ndlon, lat, lon)
            alpha = np.radians(qdr - self.ndcrs)
            base = 30. * (self.lat1 - self.lat0)
            x = dist * np.sin(alpha) / base * self.height + self.width / 2
            y = -dist * np.cos(alpha) / base * self.height + self.height / 2

        return np.rint(x), np.rint(y)


    def xy2ll(self, x, y):
        lat = self.lat0 + (self.lat1 - self.lat0) * (self.height - y) / self.height

        if self.lon1 > self.lon0:
            # Normal case
            lon = self.lon0 + (self.lon1 - self.lon0) * x / self.width

        else:
            # Wrap around:
            dellon = 360. + self.lon1 - self.lon0
            xlon = self.lon0 + x / self.width
            lon = xlon - 360. * (lon > 180.)

        # Convert pixel x,y to lat/lan [deg]
        return lat, lon


    def onradar(self, lat, lon):
        """Return boolean (also numpy array) with on or off radar screen (range check)"""

        # Radar mode
        if not self.swnavdisp:
            # Normal case
            if self.lon1 > self.lon0:
                sw = (lat > self.lat0) * (lat < self.lat1) * \
                     (lon > self.lon0) * (lon < self.lon1) == 1

            # Wrap around:
            else:
                sw = (lat > self.lat0) * (lat < self.lat1) * \
                     ((lon > self.lon0) + (lon < self.lon1)) == 1
        # Else NAVDISP mode
        else:
            base = 30. * (self.lat1 - self.lat0)
            dist = geo.latlondist(self.ndlat, self.ndlon, lat, lon) / nm
            sw = dist < base

        return sw

    @stack.command(annotations='float/txt', brief='ZOOM IN/OUT/factor')
    def zoom(self, factor):
        ''' ZOOM: Zoom in and out in the radar view. 
        
            Arguments:
            - factor: IN/OUT to zoom in/out by a factor sqrt(2), or
                      'factor' to set zoom to specific value.
        '''
        oldvalues = self.lat0, self.lat1, self.lon0, self.lon1
        absolute = True
        if isinstance(factor, str):
            absolute = False
            if factor == 'IN':
                factor = 1.4142135623730951
            elif factor == 'OUT':
                factor = 0.7071067811865475
            else:
                return False, f'ZOOM: argument {factor} not recognised'

        # Zoom factor: 2.0 means halving the display size in degrees lat/lon
        # ZOom out with e.g. 0.5

        ctrlat = (self.lat0 + self.lat1) / 2.
        if not absolute:
             dellat2 = 0.5*(self.lat1 - self.lat0) / factor
        else:
             dellat2 = 1.0/factor

        self.lat0 = ctrlat - dellat2
        self.lat1 = ctrlat + dellat2

        # Normal case
        if self.lon1 > self.lon0:
            ctrlon = (self.lon0 + self.lon1) / 2.
            dellon2 = (self.lon1 - self.lon0) / 2. / factor

        # Wrap around
        else:
            ctrlon = (self.lon0 + self.lon1 + 360.) / 2
            dellon2 = (360. + self.lon1 - self.lon0) / 2. / factor

        if absolute:
            dellon2 = dellat2 * self.width /    \
                (self.height * cos(radians(ctrlat)))

        # Wrap around
        self.lon0 = (ctrlon - dellon2 + 180.) % 360. - 180.
        self.lon1 = (ctrlon + dellon2 + 180.) % 360. - 180.

        # Avoid getting out of range
        if self.lat0 < -90 or self.lat1 > 90.:
            self.lat0, self.lat1, self.lon0, self.lon1 = oldvalues

        self.redrawradbg = True
        self.navsel = ()
        self.satsel = ()
        self.geosel = ()

        return True

    @stack.command(annotations="pandir/latlon", brief="PAN latlon/acid/airport/waypoint/LEFT/RIGHT/UP/DOWN")
    def pan(self, *args):
        "Pan screen (move view) to a waypoint, direction or aircraft"
        lat, lon = args

        # Maintain size & avoid getting out of range
        dellat2 = (self.lat1 - self.lat0) * 0.5
        self.ctrlat = max(min(lat, 90. - dellat2), dellat2 - 90.)

        # Allow wrap around of longitude
        dellon2 = dellat2 * self.width /   \
                                    (self.height * cos(radians(self.ctrlat)))
        self.ctrlon = (lon + 180.) % 360 - 180.

        # Update edge coordinates
        self.lat0 = self.ctrlat - dellat2
        self.lat1 = self.ctrlat + dellat2
        self.lon0 = (self.ctrlon - dellon2 + 180.) % 360. - 180.
        self.lon1 = (self.ctrlon + dellon2 + 180.) % 360. - 180.

        # Redraw background
        self.redrawradbg = True
        self.navsel = ()
        self.satsel = ()
        self.geosel = ()

        # Use new centre lat/lon to update reference position
        bs.ref.lat = self.ctrlat
        bs.ref.lon = self.ctrlon

        return True

    def fullscreen(self, switch):  # full screen switch
        """Switch to (True) /from (False) full screen mode"""

        # Reset screen
        pg.display.quit()
        pg.display.init()

        di = pg.display.Info()

        pg.display.set_caption("BlueSky Open ATM Simulator (F11 = Full Screen)",
                               "BlueSky")
        iconpath = bs.resource("graphics/icon.gif")
        iconbmp = pg.image.load(iconpath)
        pg.display.set_icon(iconbmp)

        if switch:
            # Full screen mode
            self.width = di.current_w
            self.height = di.current_h
            reso = (self.width, self.height)
            self.win = pg.display.set_mode(reso, pg.FULLSCREEN | pg.HWSURFACE)
        else:
            # Windowed
            self.height = min(self.height, int(di.current_h * 90 / 100))
            self.width = min(self.width, int(di.current_w * 90 / 100))
            reso = (self.width, self.height)
            self.win = pg.display.set_mode(reso)

        # Adjust scaling
        dellat = self.lat1 - self.lat0
        avelat = (self.lat0 + self.lat1) / 2.

        dellon = dellat * self.width / (self.height * cos(radians(avelat)))

        self.lon1 = self.lon0 + dellon

        self.radbmp = self.win.copy()

        # Force redraw and reselect
        self.redrawradbg = True
        self.satsel = ()
        self.navsel = ()
        self.geosel = ()

        return

    def savescreen(self):
        """Save a screenshoot"""
        now=datetime.datetime.now()
        date="%s-%s-%s" % (now.year, now.month, now.day)
        time="time=%sh %sm %ss" % (now.hour, now.minute, now.second)
        num=1
        while self.session == "new":
            self.folder="./screenshots/"+date+"-session-"+str(num)
            if os.path.exists(self.folder):
                num+=1
            else:
                os.makedirs(self.folder)
                self.session=num
        self.screenshotname=self.folder+"/"+time+".bmp"
        self.screenshot=True


    def ltopix_eq(self,lat):
        """
        Latitude to pixel conversion. Compute how much pixels a
        degree in latlon is in longitudinal direction.
        """

        pwidth=self.width
        lwidth=self.lon1-self.lon0

        return int(lat/lwidth*pwidth)


    def dtopix_eq(self,dist):
        """
        Distance to pixel conversion. Compute how much pixels a
        meter is in longitudinal direction
        """
        lat=dist/111319.
        return self.ltopix_eq(lat)

    @stack.commandgroup(annotations='txt,color', aliases=('COLOR', 'COL'))
    def colour(self, name, r, g, b):
        ''' Set custom colour for aircraft or shape. '''
        if areafilter.hasArea(name):
            idx = self.objname.index(name)
            self.objcolor[idx] = (r, g, b)
        else:
            return False, 'No object found with name ' + name

        self.redrawradbg = True  # redraw background
        return True

    def objappend(self,itype,name,data):
        """Add user defined objects"""
        if not kwargs:
            return self.objdel()

        if name not in self.objname:
            self.objname.append(name)
            self.objtype.append(kwargs['shape'])
            if self.objtype[-1]==1:
                self.objtype[-1]="LINE" # Convert to string

            self.objcolor.append(kwargs.get('color', cyan))
            self.objdata.append(kwargs['coordinates'])
        else:
            idx = self.objname.index(name)
            self.objtype[idx] = kwargs.get('shape', self.objtype[idx])
            self.objcolor[idx] = kwargs.get('color', self.objcolor[idx])
            self.objdata[idx] = kwargs.get('coordinates', self.objdata[idx])

        self.redrawradbg = True  # redraw background

        return

    def objdel(self):
        """Add user defined objects"""
        self.objname     = []
        self.objtype     = []
        self.objcolor    = []
        self.objdata     = []
        self.redrawradbg = True  # redraw background
        return

    def showroute(self, acid):  # Toggle show route for an aircraft id
        if self.acidrte == acid:
            self.acidrte = ""  # Click twice on same: route disappear
        else:
            self.acidrte = acid  # Show this route
        return True


    def addnavwpt(self,name,lat,lon): # Draw new navdb waypoint
        # As in pygame navdb has already updated data, simply redraw background
        self.wpswbmp.append(False) # Add cell to buffer
        self.wplabel.append(0) # Add cell to buffer
        self.redrawradbg = True  # redraw background
        return

    def getviewctr(self):
        return (self.ctrlat, self.ctrlon)

    def getviewbounds(self): # Return current viewing area in lat, lon
        return self.lat0, self.lat1, self.lon0, self.lon1

    def drawradbg(self): # redraw radar background
        self.redrawradbg = True
        return

    @stack.command
    def feature(self,sw,arg=""):
        ''' Switch on/off elements and background of map/radar view 
        
            Usage:
                SWRAD GEO/GRID/APT/VOR/WPT/LABEL/ADSBCOVERAGE/TRAIL/POLY [dt]/[value]
        '''
        # Switch/toggle/cycle radar screen features e.g. from SWRAD command
        # Coastlines
        if sw == "GEO":
            self.swgeo = not self.swgeo


        # FIR boundaries
        elif sw == "FIR":
            self.swfir = not self.swfir

        # Airport: 0 = None, 1 = Large, 2= All
        elif sw == "APT":
            self.apsw = (self.apsw + 1) % 3
            if not (arg== ""):
                self.apsw = int(cmdargs[2])
            self.navsel = []

        # Waypoint: 0 = None, 1 = VOR, 2 = also WPT, 3 = Also terminal area wpts
        elif sw == "VOR" or sw == "WPT" or sw == "WP" or sw == "NAV":
            self.wpsw = (self.wpsw + 1) % 4
            if not (arg== ""):
                self.wpsw = int(arg)
            self.navsel = []

        # Satellite image background on/off
        elif sw == "SAT":
            self.swsat = not self.swsat

        elif sw[:4] == "ADSB":
            self.swAdsbCoverage = not self.swAdsbCoverage

        # Traffic labels: cycle nr of lines 0,1,2,3
        elif sw[:3] == "LAB":  # Nr lines in label
            self.swlabel = (self.swlabel + 1) % 4
            if not (arg == ""):
                self.swlabel = int(arg)

        else:
            self.redrawradbg = False
            return False # switch not found

        self.redrawradbg = True
        return True # Success

    def show_file_dialog(self):
        return opendialog()
    
    @stack.command
    def symbol(self):
        self.swsep = not self.swsep
        return True

    def show_cmd_doc(self, cmd=''):
        # Show documentation on command
        if not cmd:
            cmd = 'Command-Reference'
        curdir = os.getcwd()
        os.chdir("html")
        htmlfile = cmd.lower()+".html"
        if os.path.isfile(htmlfile):
            try:
                subprocess.Popen(htmlfile,shell=True)
            except:
                os.chdir(curdir)
                return False,"Opening "+htmlfile+" failed."
        else:
            os.chdir(curdir)
            return False,htmlfile+" is not yet available, try HELP PDF or check the wiki on Github."

        os.chdir(curdir)
        return True,"HTML window opened"



""" BlueSky plugin template. The text you put here will be visible
    in BlueSky as the description of your plugin. """

# flake8: noqa

# Developed by Julia and Jerom, superteam!

# Import the global bluesky objects. Uncomment the ones you need
from bluesky import stack, traf, sim  #, settings, navdb, traf, sim, scr, tools
from bluesky.tools.aero import kts, ft, fpm, vcas2tas
from bluesky.tools import geo
import numpy as np


# Globals
commandnames = []
# Make a list of testing functions
fnumber = 0
teststart = True
timer = 0.0


### Initialization function of your plugin. Do not change the name of this
### function, as it is the way BlueSky recognises this file as a plugin.
def init_plugin():
    global fnumber, commandnames, teststart

    # Run the time
    stack.stack("OP")
    stack.stack("FF")
    
    commandnames = list(stack.cmddict.keys())
    # Make a list of testing functions
    fnumber = 0
    teststart = True


    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'STACKCHECK',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim',
        'update':          update,
        'preupdate':       preupdate
        }

    stackfunctions = {
        # The command name for your function
        'MYFUN': [
            # A short usage string. This will be printed if you type HELP <name> in the BlueSky console
            'MYFUN ON/OFF',

            # A list of the argument types your function accepts. For a description of this, see ...
            '[onoff]',

            # The name of your function in this plugin
            myfun,

            # a longer help text of your function.
            'Print something to the bluesky console based on the flag passed to MYFUN.']
    }

    # init_plugin() should always return these two dicts.
    return config, stackfunctions


### Periodic update functions that are called by the simulation. You can replace
### this by anything, so long as you communicate this in init_plugin

def update():
    global commandnames, fnumber, teststart, timer
    if fnumber < len(commandnames):
        # form function name
        func = "test_"+ commandnames[fnumber].lower()
        timer = sim.simt

        # if function exists
        if func in globals().keys():
            # run the test and get result
            result = globals()[func](teststart)
            teststart = False
            if result != None:
                stack.stack("ECHO "+commandnames[fnumber]+" DOES %sWORK" % ("" if result else "NOT ") )
                # Prepare for next test function
                fnumber += 1
                teststart = True
                # reset traffic to get rid of unnecessary crap
                traf.reset()

        else:
            # if function does not exist - keep looping
            fnumber += 1
            #stack.stack("ECHO "+func+" DOES NOT EXIST")
    else:
        stack.stack("RESET")
        stack.stack("PAUSE")
        stack.stack("ECHO FINISHED PERFORMING STACKCHECK")
        stack.stack("PLUGIN REMOVE STACKCHECK")

def test_addwpt(start):
    global starttime, timer
    timelimit = 120
    if start:
        # Create a waypoint which does not lie on the aircraft path
        stack.stack("CRE KLM10 B747 52 4 000 FL99 150")
        stack.stack("KLM10 ADDWPT 52 4.1")
        starttime = timer
    else:
        # When the aircraft reaches the waypoint, return success
        _, d = geo.qdrdist(52,4.1,traf.lat[0],traf.lon[0])
        closeenough = d < 0.2 # d is measured in nm
        if timer-starttime>timelimit or closeenough:
            return closeenough

def test_after(start):
    global starttime, timer
    timelimit = 240
    lowertimelimit = 120
    if start:
        # Create a waypoint which does not lie on the aircraft path
        # And create a next waypoint to reach after that
        stack.stack("CRE KLM10 B747 52 4 000 FL99 150")
        stack.stack("DEFWPT TESTWP 52 4.1")
        stack.stack("KLM10 ADDWPT TESTWP")
        stack.stack("KLM10 AFTER TESTWP ADDWPT 51.9 4.1")
        starttime = timer
    else:
        # When the aircraft reaches the waypoint, return success
        _, d = geo.qdrdist(51.9,4.1,traf.lat[0],traf.lon[0])
        closeenough = d < 0.2 # d is measured in nm
        if timer-starttime>timelimit or closeenough:
            return closeenough and timer-starttime>lowertimelimit

def test_alt(start):
    global starttime, timer
    timelimit = 60.
    fl = 100
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 FL99 250")
        stack.stack("KLM10 ALT FL%s" %fl)
        starttime = timer
    else:
        closeenough = (fl*100 - 10) * ft <= traf.alt[0] < (fl*100 + 10) * ft
        if (timer - starttime) > timelimit or closeenough:
            return closeenough

def test_before(start):
    global starttime, timer
    timelimit = 240
    if start:
        # Create a waypoint which does not lie on the aircraft path
        # And create a next waypoint to reach after that
        stack.stack("CRE KLM10 B747 52 4 000 FL99 150")
        stack.stack("DEFWPT WP1 52.0 4.1")
        stack.stack("DEFWPT WP2 52.1 4.2")
        stack.stack("DEFWPT WP3 52.0 4.3")
        stack.stack("KLM10 ADDWPT WP1")
        stack.stack("KLM10 AFTER WP1 ADDWPT WP3")
        stack.stack("KLM10 BEFORE WP3 ADDWPT WP2")
        starttime = timer
    else:
        # When the aircraft reaches waypoint 3 within the time limit return success
        _, d = geo.qdrdist(52.1 ,4.2 ,traf.lat[0],traf.lon[0])
        closeenough = d < 4 # d is measured in nm
        if timer-starttime>timelimit or closeenough:
            return closeenough

def test_cre(start):
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 59 250")
    else:
        return traf.ntraf==1

def test_defwpt(start):
    global starttime, timer
    timelimit = 120
    if start:
        # Create a waypoint which does not lie on the aircraft path using defwpt
        stack.stack("CRE KLM10 B747 52 4 000 FL99 150")
        stack.stack("DEFWPT TESTWP 52 4.1")
        stack.stack("KLM10 ADDWPT TESTWP")
        starttime = timer
    else:
        # When the aircraft reaches the waypoint, return success
        _, d = geo.qdrdist(52,4.1,traf.lat[0],traf.lon[0])
        closeenough = d < 0.2 # d is measured in nm
        if timer-starttime>timelimit or closeenough:
            return closeenough

def test_del(start):
    global starttime, timer
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 59 250")
        stack.stack("CRE KLM11 B747 53 4 000 59 250")
        stack.stack("CRE KLM12 B747 54 4 000 59 250")
        starttime = timer
    elif traf.ntraf==3 and timer-starttime<5:
        stack.stack("DEL KLM10")
    else:
        return traf.ntraf==2

def test_delwpt(start):
    global starttime, timer, wp2reached
    timelimit = 240
    if start:
        # Create a waypoint which does not lie on the aircraft path
        # And create a next waypoint to reach after that
        stack.stack("CRE KLM10 B747 52 4 000 FL99 150")
        stack.stack("DEFWPT WPDEL1 52.0 4.1")
        stack.stack("DEFWPT WPDEL2 52.1 4.2")
        stack.stack("DEFWPT WPDEL3 52.0 4.3")
        stack.stack("KLM10 ADDWPT WPDEL1")
        stack.stack("KLM10 ADDWPT WPDEL2")
        stack.stack("KLM10 ADDWPT WPDEL3")
        stack.stack("KLM10 DELWPT WPDEL2")
        starttime = timer
        wp2reached = False
    else:
        _, d2 = geo.qdrdist(52.1 ,4.2 ,traf.lat[0],traf.lon[0])
        if d2 < 1:
            wp2reached = True
        # When the aircraft reaches waypoint 3 within the time limit return success
        _, d = geo.qdrdist(52.0 ,4.3 ,traf.lat[0],traf.lon[0])
        wp3reached = d < 1 # d is measured in nm
        if timer-starttime>timelimit or wp3reached:
            return wp3reached and not wp2reached

def test_dest(start):
    dest = 'EHAM'
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 FL100 250")
        stack.stack("KLM10 DEST %s"%dest)
    else:
        return traf.ap.dest[0] == dest

def test_direct(start):
    global starttime, timer, wp1reached
    timelimit = 350
    if start:
        # Create a waypoint which does not lie on the aircraft path
        # And create the waypoint to direct to
        stack.stack("CRE KLM10 B747 52 4 000 FL99 150")
        stack.stack("DEFWPT WPDIR1 52.2 4.1")
        stack.stack("DEFWPT WPDIR2 52.2 3.9")
        stack.stack("KLM10 ADDWPT WPDIR1")
        stack.stack("KLM10 ADDWPT WPDIR2")
        stack.stack("KLM10 DIRECT WPDIR2")
        starttime = timer
        wp1reached = False
    else:
        # If close to WP1, save that WP1 is reached
        _, d1 = geo.qdrdist(52.2 , 4.1 ,traf.lat[0],traf.lon[0])
        if d1 < 1:
            wp1reached = True

        # When the aircraft reaches waypoint 2 within the time limit, return success
        _, d2 = geo.qdrdist(52.2 , 3.9 ,traf.lat[0],traf.lon[0])
        closeenough = d2 < 1 # d is measured in nm
        if timer-starttime>timelimit or closeenough:
            return closeenough and not wp1reached

def test_eng(start):
    """ at a moment the engine change DOES result in performance coefficient change
    in accordance with a new engine, but the engine id/type is not changed
    """
    global starttime, timer
    timelimit = 5.  # seconds
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 FL100 250")
        starttime = timer
    # elif timer - starttime < timelimit:
    #     print(traf.perf.engines)
    #     print("eng type before change", traf.perf.etype[0])
    #     stack.stack("ENG KLM10 RB211-22B")
    #     print("eng type after change", traf.perf.etype[0])
    else:
        return False

def test_hdg(start):
    global starttime, timer
    timelimit = 70.  # seconds
    new_heading = 170
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 FL100 250")
        stack.stack("KLM10 HDG %s"%new_heading)
        starttime = timer
    else:
        closeenough = (new_heading - 1) <= traf.hdg[0] < (new_heading + 1)
        if (timer - starttime) > timelimit or closeenough:
            return closeenough

def test_lnav(start):
    global starttime, timer, detourmade
    timelimit = 150
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 59 250")
        stack.stack("CRE KLM11 B747 52 5.1 000 59 250")
        stack.stack("KLM10 ADDWPT 52.2 4")
        stack.stack("KLM11 ADDWPT 52.2 5.1")
        stack.stack("KLM10 ADDWPT 52.2 3.5")
        stack.stack("KLM11 ADDWPT 52.2 5.6")
        stack.stack("KLM10 LNAV OFF")
        stack.stack("KLM10 LNAV ON")
        stack.stack("KLM11 LNAV OFF")
        starttime = timer
    elif timer-starttime > timelimit:
        lnavon = traf.hdg[0]>=265
        lnavoff = traf.hdg[1]<= 5 or traf.hdg[1] >=355
        return lnavon and lnavoff

def test_mcre(start):
    ac_type = 'B747'
    alt = 100  # FL
    spd = 250  # CAS in knots
    ades = 'EHAM'  # airport of destination is not checked yet
    if start:
        stack.stack("MCRE 11")
    elif traf.ntraf == 11:
        traf.reset()
        stack.stack("MCRE 5 %s FL%s %s %s" % (ac_type, alt, spd, ades))
    else:

        close_enough_alt = np.logical_and(traf.alt <= (alt * 100 * ft + 10),
                                          traf.alt >= (alt * 100 * ft - 10))

        close_enough_spd = np.logical_and(traf.cas <= (spd * kts + 10),
                                          traf.cas >= (spd * kts - 10))

        return (traf.ntraf == 5 and len(traf.type)==5 and
                np.all(close_enough_alt) and np.all(close_enough_spd))

def test_move(start):
    # set new values for a MOVE command 
    new_lat, new_lon = 52, 8
    new_alt = 200  # FL
    new_hdg = 180  # degrees
    new_spd = 300  # CAS in knots
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 FL100 250")
        stack.stack("MOVE KLM10 52 8 FL200 180 300")
    else:
        # check whether variables are within the reasonable ranges from the set values
        close_enough_lat = new_lat - 0.1 <= traf.lat[0] <= new_lat + 0.1
        close_enough_lon = new_lon - 0.1 <= traf.lon[0] <= new_lon + 0.1
        close_enough_hdg = new_hdg - 1 <= traf.hdg[0] <= new_hdg + 1
        close_enough_alt = new_alt * 100 * ft - 10 <= traf.alt[0] <= new_alt * 100 * ft + 10
        close_enough_spd = new_spd * kts - 5 <= traf.cas[0] <= new_spd * kts + 5
        return close_enough_lat and close_enough_lon and close_enough_hdg and close_enough_alt and close_enough_spd

def test_orig(start):
    orig = 'EHAM'
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 FL100 250")
        stack.stack("KLM10 ORIG %s"%orig)
    else:
        return traf.ap.orig[0] == orig

def test_spd(start):
    global starttime, timer
    timelimit = 60.
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 59 250")
        stack.stack("KLM10 SPD 300")
        starttime = timer
    else:
        closeenough = 295 *kts <= traf.cas[0] < 305*kts
        if timer-starttime > timelimit or closeenough:
            return closeenough

def test_vnav(start):
    global starttime, timer, detourmade
    timelimit = 60
    if start:
        stack.stack("CRE KLM10 B747 52 4 000 FL59 250")
        stack.stack("CRE KLM11 B747 52 5.1 000 FL59 250")
        stack.stack("KLM10 ADDWPT 52.1 4 FL65")
        stack.stack("KLM11 ADDWPT 52.1 5.1 FL65")
        stack.stack("KLM10 VNAV OFF")
        stack.stack("KLM10 VNAV ON")
        stack.stack("KLM11 VNAV OFF")
        starttime = timer
    elif timer-starttime > timelimit:
        vnavon  = traf.alt[0] >= 6200 *ft
        vnavoff = 5800*ft < traf.alt[1] < 6000 *ft
        return vnavon and vnavoff

def test_vs(start):
    global starttime, timer
    timelimit = 10.  # seconds
    new_vs = 2000  # fpm
    init_alt, new_alt = 100, 200
    if start:
        # Create an aircraft that should perform a vertical manoever
        stack.stack("CRE KLM10 B747 52 4 000 FL%s 250"%init_alt)
        stack.stack("KLM10 ALT FL%s"%new_alt)
        stack.stack("KLM10 VS %s" % new_vs)
        starttime = timer
    else:
        # Check if the aircraft adopted the new vs
        closeenough = (new_vs * fpm - 10) <= traf.vs[0] < (new_vs * fpm + 10)
        if (timer - starttime) > timelimit or closeenough:
            return closeenough

def test_wind(start):
    global starttime, timer
    timelimit = 20
    if start:
        # Create a very strong headwind
        cas = 250
        tas = vcas2tas(cas*kts,10000*ft)/kts
        stack.stack("CRE KLM10 B747 52 4 000 FL100 "+str(int(cas)))
        # Mind that wind is defined in the direction that it is coming from
        stack.stack("WIND 52 4 FL100 000 "+str(int(tas)))
        starttime = timer
    elif timer-starttime>timelimit:
        # If the aircraft did not move 
        _, d = geo.qdrdist(52,4,traf.lat[0],traf.lon[0])
        return d < 0.1 # d is measured in nm

def preupdate():
    pass

### Other functions of your plugin
def myfun(flag=True):
    return True, 'My plugin received an o%s flag.' % ('n' if flag else 'ff')



''' BlueSky scenario recorder. '''
import math
from pathlib import Path

import bluesky as bs
from bluesky.tools.aero import kts, ft, fpm, tas2cas, density,casormach2tas
from bluesky.tools.misc import tim2txt
from bluesky.stack.cmdparser import command, commandgroup

# When SAVEIC is used, we will also have a recording scenario file handle
savefile = None  # File object of recording scenario file
defexcl = [
    "PAN",
    "ZOOM",
    "HOLD",
    "POS",
    "INSEDIT",
    "SAVEIC",
    "QUIT",
    "PCALL",
    "CALL",
    "PLOT",
    "CALC",
    "FF",
    "IC",
    "OP",
    "HOLD",
    "RESET",
    "MCRE",
    "CRE",
    "TRAFGEN",
    "LISTRTE",
]  # Commands to be excluded, default
# Note (P)CALL is always excluded! Commands in called file are saved explicitly
saveexcl = defexcl
# simt time of moment of SAVEIC command, 00:00:00.00 in recorded file
saveict0 = 0.0

@commandgroup
def saveic(filename: 'word' = ''):
    """ Save the current traffic realization in a scenario file. """
    global savefile, saveict0

    # No args? Give current status
    if not filename:
        if savefile is None:
            return False, "SAVEIC is not running"
        else:
            return True, "SAVEIC is already on\n" + "File: " + savefile.name

    # If recording is already on, give message
    if savefile is not None:
        return False, "SAVEIC is already on\n" + "Savefile:  " + savefile.name

    # Add extension .scn if not already present
    filename = Path(filename).with_suffix('.scn')
    # If it is with path don't touch it, else add path
    if not filename.is_absolute():
        filename = bs.resource(bs.settings.scenario_path) / filename

    try:
        f = open(filename, "w")
    except:
        return False, "Error writing to file"

    # Write files
    timtxt = "00:00:00.00>"  # Current time will be zero
    saveict0 = bs.sim.simt

    for i in range(bs.traf.ntraf):
        # CRE acid,type,lat,lon,hdg,alt,spd
        cmdline = (f'CRE {bs.traf.id[i]},{bs.traf.type[i]},{bs.traf.lat[i]},'
                   f'{bs.traf.lon[i]},{bs.traf.trk[i]},{bs.traf.alt[i] / ft},'
                   f'{tas2cas(bs.traf.tas[i], bs.traf.alt[i]) / kts}')

        f.write(timtxt + cmdline + "\n")


        # Autopilot commands
        # Altitude & VS
        if abs(bs.traf.alt[i] - bs.traf.ap.alt[i]) > 10.0:
            cmdline = "ALT " + bs.traf.id[i] + \
                "," + repr(bs.traf.ap.alt[i] / ft)
            f.write(timtxt + cmdline + "\n")

            # VS acid,vs
            if abs(bs.traf.vs[i]) > 0.05:  # 10 fpm dead band
                if abs(bs.traf.ap.vs[i]) > 0.05:
                    vs_ = bs.traf.ap.vs[i] / fpm
                else:
                    vs_ = bs.traf.vs[i] / fpm

                cmdline = f'VS {bs.traf.id[i]},{vs_}'
                f.write(timtxt + cmdline + "\n")

        # Heading as well when heading select
        delhdg = (bs.traf.hdg[i] - bs.traf.ap.trk[i] + 180.0) % 360.0 - 180.0
        if abs(delhdg) > 0.5:
            cmdline = "HDG " + bs.traf.id[i] + "," + repr(bs.traf.ap.trk[i])
            f.write(timtxt + cmdline + "\n")

        # Speed select mode? => Record a speed command
        aptas = casormach2tas(bs.traf.selspd[i],bs.traf.ap.spd[i])
        delspd = aptas - bs.traf.tas[i]

        if abs(delspd) > 0.5*kts: # difference equal more than 1 knot (rounded), so 0.5* 0.514444 m/s
            if not(0.001<bs.traf.selspd[i]<1.0): # Check for Mach
                spdcmd = round(bs.traf.selspd[i]/kts,2)
            else:
                spdcmd = bs.traf.selspd[i]

            cmdline = "SPD " + bs.traf.id[i] + \
                "," + repr(spdcmd)
            f.write(timtxt + cmdline + "\n")

        # DEST acid,dest-apt
        if bs.traf.ap.dest[i] != "":
            cmdline = "DEST " + bs.traf.id[i] + "," + bs.traf.ap.dest[i]
            f.write(timtxt + cmdline + "\n")

        # ORIG acid,orig-apt
        if bs.traf.ap.orig[i] != "":
            cmdline = "ORIG " + bs.traf.id[i] + "," + bs.traf.ap.orig[i]
            f.write(timtxt + cmdline + "\n")

        # Route with ADDWPT
        route = bs.traf.ap.route[i]
        for iwp in range(route.nwp):
            # dets and orig al already done, skip them here
            if iwp == 0 and route.wpname[iwp] == bs.traf.ap.orig[i]:
                continue

            if iwp == route.nwp - 1 and route.wpname[iwp] == bs.traf.ap.dest[i]:
                continue

            # add other waypoints
            cmdline = "ADDWPT " + bs.traf.id[i] + " "
            wpname = route.wpname[iwp]
            if wpname[: len(bs.traf.id[i])] == bs.traf.id[i]:
                wpname = repr(route.wplat[iwp]) + "," + repr(route.wplon[iwp])
            cmdline = cmdline + wpname + ","

            if route.wpalt[iwp] >= 0.0:
                cmdline = cmdline + repr(route.wpalt[iwp] / ft) + ","
            else:
                cmdline = cmdline + ","

            if route.wpspd[iwp] >= 0.0:
                if route.wpspd[iwp] > 1.0:
                    cmdline = cmdline + repr(route.wpspd[iwp] / kts)
                else:
                    cmdline = cmdline + repr(route.wpspd[iwp])

            f.write(timtxt + cmdline + "\n")

    # Saveic: save file
    savefile = f
    return True


@saveic.subcommand(name='EXCEPT')
def setexcept(*commands: 'txt'):
    ''' Indicate commands that need to be omitted by SAVEIC. '''
    global saveexcl
    if not commands:
        return True, "EXCEPT is now: " + " ".join(saveexcl)

    if 'NONE' in commands:
        # All commands should be saved, only skip the bare minimum
        saveexcl = ["INSEDIT", "SAVEIC"]
    else:
        saveexcl = command
    return True


@saveic.subcommand(name='CLOSE')
def saveclose():
    ''' Reset recorder. '''
    global savefile
    if savefile is not None:
        savefile.close()

    savefile = None
    return True


def savecmd(cmd, line):
    ''' Save command line to file if SAVEIC is turned on. '''
    if savefile is None or cmd in saveexcl:
        return

    # Write in format "HH:MM:SS.hh>
    timtxt = tim2txt(bs.sim.simt - saveict0)
    savefile.write(f'{timtxt}>{line}\n')


def reset():
    ''' Reset SAVEIC recorder: close file and reset excluded command list. '''
    global saveexcl
    saveclose()
    saveexcl = defexcl


"""A plugin for playing air traffic from the OpenSky Network.

The plugin gets current traffic from the OpenSky Network and makes aircraft
move.

The OpenSky Python API allows for a number of unauthenticated requests. If you
feed the network, you are allowed an unlimited number of requests for the data
you feed. If you set `opensky_user` and `opensky_password` in your settings.cfg
file and the latest request you send fails, the program falls back to the data
from your sensors.

Xavier Olive, 2018
Joost Ellerbroek, 2018
"""
import time
import pickle
import requests
import numpy as np

from bluesky import stack, settings, traf
from bluesky.core import Entity, timed_function
from bluesky.tools import cachefile
settings.set_variable_defaults(opensky_user=None, opensky_password=None,
                               opensky_ownonly=False)

# Globals
actypedb_version = 'v20180126'
reader = None
actypes = dict()

def init_plugin():
    global reader
    reader = OpenSkyListener()

    # Load aircraft type database
    get_actypedb()

    config = {
        'plugin_name': 'OPENSKY',
        'plugin_type': 'sim'
    }

    return config

def get_actypedb():
    ''' Get aircraft type database from cache or web. '''
    global actypes
    with cachefile.openfile('actypedb.p', actypedb_version) as cache:
        try:
            actypes = cache.load()
        except (pickle.PickleError, cachefile.CacheError) as e:
            print(e.args[0])
            import io, zipfile
            f = requests.get('https://junzisun.com/adb/download')
            with zipfile.ZipFile(io.BytesIO(f.content)) as zfile:
                with zfile.open('aircraft_db.csv', 'r') as dbfile:
                    actypes = dict([line.decode().split(',')[0:3:2] for line in dbfile])

            cache.dump(actypes)


class OpenSkyListener(Entity):
    def __init__(self):
        super().__init__()
        if settings.opensky_user:
            self._auth = (settings.opensky_user, settings.opensky_password)
        else:
            self._auth = ()
        self._api_url = "https://opensky-network.org/api"
        self.connected = False

        with self.settrafarrays():
            self.upd_time = np.array([])
            self.my_ac = np.array([], dtype=bool)

    def create(self, n=1):
        super().create(n)
        # Store creation time of new aircraft
        self.upd_time[-n:] = time.time()
        self.my_ac[-n:] = False

    def get_json(self, url_post, params=None):
        r = requests.get(self._api_url + url_post, auth=self._auth, params=params)
        if r.status_code == 200:
            return r.json()

        # "Response not OK. Status {0:d} - {1:s}".format(r.status_code, r.reason)
        return None

    def get_states(self, ownonly=False):
        url_post = '/states/{}'.format('own' if ownonly else 'all')
        states_json = self.get_json(url_post)
        if states_json is not None:
            return list(zip(*states_json['states']))
        return None

    @timed_function(name='OPENSKY', dt=6.0)
    def update(self):
        if not self.connected:
            return

        # t1 = time.time()
        if settings.opensky_ownonly:
            states = self.get_states(ownonly=True)
            if states is None:
                return
        else:
        # Get states from OpenSky. If all states fails try getting own states only.
            states = self.get_states()
            if states is None:
                if self.authenticated:
                    states = self.get_states(ownonly=True)
                if states is None:
                    return

        # Current time
        curtime = time.time()

        # States contents:
        icao24, acid, orig, time_pos, last_contact, lon, lat, geo_alt, on_gnd, \
            spd, hdg, vspd, sensors, baro_alt, squawk, spi, pos_src = states[:17]

        # Relevant params as numpy arrays
        lat = np.array(lat, dtype=np.float64)
        lon = np.array(lon, dtype=np.float64)
        alt = np.array(baro_alt, dtype=np.float64)
        hdg = np.array(hdg, dtype=np.float64)
        vspd = np.array(vspd, dtype=np.float64)
        spd = np.array(spd, dtype=np.float64)
        acid = np.array([i.strip() for i in acid], dtype=np.str_)
        icao24 = np.array(icao24, dtype=np.str_)

        idx = np.array(traf.id2idx(acid))

        # Split between already existing and new aircraft
        newac = idx == -1
        other = np.logical_not(newac)

        # Filter out invalid entries
        valid = np.logical_not(np.logical_or.reduce(
            [np.isnan(x) for x in [lat, lon, alt, hdg, vspd, spd]]))
        newac = np.logical_and(newac, valid)
        other = np.logical_and(other, valid)
        n_new = np.count_nonzero(newac)
        n_oth = np.count_nonzero(other)

        # t2 = time.time()

        # Create new aircraft
        if n_new:
            actype = [actypes.get(str(i), 'B744') for i in icao24[newac]]
            for j in range(n_new):
                traf.cre(n_new, acid[newac][j], actype[j], lat[newac][j], lon[newac][j],\
                         hdg[newac][j], alt[newac][j], spd[newac][j])
            self.my_ac[-n_new:] = True

        # t3 = time.time()

        # Update the rest
        if n_oth:
            traf.move(idx[other], lat[other], lon[other], alt[other], hdg[other], \
                      spd[other], vspd[other])
            self.upd_time[idx[other]] = curtime

        # t4 = time.time()

        # remove aircraft with no message for less than 1 minute
        # opensky already filters
        delidx = np.where(np.logical_and(self.my_ac, curtime - self.upd_time > 10))[0]
        if len(delidx) > 0:
            traf.delete(delidx)

        # t5 = time.time()
        # print('req={}, mod={}, cre={}, mov={}, del={}, ncre={}, nmov={}, ndel={}'.format(curtime-t1, t2-curtime, t3-t2, t4-t3, t5-t4, n_new, n_oth, len(delidx)))

    @stack.command(name='OPENSKY')
    def toggle(self, flag:bool=None):
        ''' Use the OpenSky plugin as a data source for traffic. '''
        if flag is None:
            return True, f'OpenSky is currently {"" if self.connected else "not"} connected.'
        self.connected = flag
        if flag:
            stack.stack('OP')
            return True, 'Connecting to OpenSky'
        return True, 'Stopping the requests'



''' Conflict resolution based on the Modified Voltage Potential algorithm. '''
import numpy as np
from bluesky import stack
from bluesky.traffic.asas import ConflictResolution


class MVP(ConflictResolution):
    ''' Conflict resolution using the Modified Voltage Potential Method. '''
    def __init__(self):
        super().__init__()
        # [-] switch to limit resolution to the horizontal direction
        self.swresohoriz = True
        # [-] switch to use only speed resolutions (works with swresohoriz = True)
        self.swresospd = False
        # [-] switch to use only heading resolutions (works with swresohoriz = True)
        self.swresohdg = False
        # [-] switch to limit resolution to the vertical direction
        self.swresovert = False

    def setprio(self, flag=None, priocode=''):
        '''Set the prio switch and the type of prio '''
        if flag is None:
            return True, "PRIORULES [ON/OFF] [PRIOCODE]" + \
                            "\nAvailable priority codes: " + \
                            "\n     FF1:  Free Flight Primary (No Prio) " + \
                            "\n     FF2:  Free Flight Secondary (Cruising has priority)" + \
                            "\n     FF3:  Free Flight Tertiary (Climbing/descending has priority)" + \
                            "\n     LAY1: Layers Primary (Cruising has priority + horizontal resolutions)" + \
                            "\n     LAY2: Layers Secondary (Climbing/descending has priority + horizontal resolutions)" + \
                            "\nPriority is currently " + ("ON" if self.swprio else "OFF") + \
                            "\nPriority code is currently: " + \
                str(self.priocode)
        options = ["FF1", "FF2", "FF3", "LAY1", "LAY2"]
        if priocode not in options:
            return False, "Priority code Not Understood. Available Options: " + str(options)
        return super().setprio(flag, priocode)

    @stack.command(name="RMETHH")
    def setresometh(self, value:'txt'=''):
        """ Processes the RMETHH command. Sets swresovert = False"""
        # Acceptable arguments for this command
        options = ["BOTH", "SPD", "HDG", "NONE", "ON", "OFF", "OF"]
        if not value:
            return True, "RMETHH [ON / BOTH / OFF / NONE / SPD / HDG]" + \
                         "\nHorizontal resolution limitation is currently " + ("ON" if self.swresohoriz else "OFF") + \
                         "\nSpeed resolution limitation is currently " + ("ON" if self.swresospd else "OFF") + \
                         "\nHeading resolution limitation is currently " + \
                ("ON" if self.swresohdg else "OFF")
        if value not in options:
            return False, "RMETH Not Understood" + "\nRMETHH [ON / BOTH / OFF / NONE / SPD / HDG]"
        else:
            if value == "ON" or value == "BOTH":
                self.swresohoriz = True
                self.swresospd = True
                self.swresohdg = True
                self.swresovert = False
            elif value == "OFF" or value == "OF" or value == "NONE":
                # Do NOT swtich off self.swresovert if value == OFF
                self.swresohoriz = False
                self.swresospd = False
                self.swresohdg = False
            elif value == "SPD":
                self.swresohoriz = True
                self.swresospd = True
                self.swresohdg = False
                self.swresovert = False
            elif value == "HDG":
                self.swresohoriz = True
                self.swresospd = False
                self.swresohdg = True
                self.swresovert = False

    @stack.command(name='RMETHV')
    def setresometv(self, value:'txt'=''):
        """ Processes the RMETHV command. Sets swresohoriz = False."""
        # Acceptable arguments for this command
        options = ["NONE", "ON", "OFF", "OF", "V/S"]
        if not value:
            return True, "RMETHV [ON / V/S / OFF / NONE]" + \
                         "\nVertical resolution limitation is currently " + \
                ("ON" if self.swresovert else "OFF")
        if value not in options:
            return False, f"RMETHV '{value}' Not Understood\nRMETHV [ON / V/S / OFF / NONE]"

        if value == "ON" or value == "V/S":
            self.swresovert = True
            self.swresohoriz = False
            self.swresospd = False
            self.swresohdg = False
        elif value == "OFF" or value == "OF" or value == "NONE":
            # Do NOT swtich off self.swresohoriz if value == OFF
            self.swresovert = False

    def applyprio(self, dv_mvp, dv1, dv2, vs1, vs2):
        ''' Apply the desired priority setting to the resolution '''

        # Primary Free Flight prio rules (no priority)
        if self.priocode == 'FF1':
            # since cooperative, the vertical resolution component can be halved, and then dv_mvp can be added
            dv_mvp[2] = dv_mvp[2] / 2.0
            dv1 = dv1 - dv_mvp
            dv2 = dv2 + dv_mvp

        # Secondary Free Flight (Cruising aircraft has priority, combined resolutions)
        if self.priocode == 'FF2':
            # since cooperative, the vertical resolution component can be halved, and then dv_mvp can be added
            dv_mvp[2] = dv_mvp[2]/2.0
            # If aircraft 1 is cruising, and aircraft 2 is climbing/descending -> aircraft 2 solves conflict
            if abs(vs1) < 0.1 and abs(vs2) > 0.1:
                dv2 = dv2 + dv_mvp
            # If aircraft 2 is cruising, and aircraft 1 is climbing -> aircraft 1 solves conflict
            elif abs(vs2) < 0.1 and abs(vs1) > 0.1:
                dv1 = dv1 - dv_mvp
            else:  # both are climbing/descending/cruising -> both aircraft solves the conflict
                dv1 = dv1 - dv_mvp
                dv2 = dv2 + dv_mvp

        # Tertiary Free Flight (Climbing/descending aircraft have priority and crusing solves with horizontal resolutions)
        elif self.priocode == 'FF3':
            # If aircraft 1 is cruising, and aircraft 2 is climbing/descending -> aircraft 1 solves conflict horizontally
            if abs(vs1) < 0.1 and abs(vs2) > 0.1:
                dv_mvp[2] = 0.0
                dv1 = dv1 - dv_mvp
            # If aircraft 2 is cruising, and aircraft 1 is climbing -> aircraft 2 solves conflict horizontally
            elif abs(vs2) < 0.1 and abs(vs1) > 0.1:
                dv_mvp[2] = 0.0
                dv2 = dv2 + dv_mvp
            else:  # both are climbing/descending/cruising -> both aircraft solves the conflict, combined
                dv_mvp[2] = dv_mvp[2]/2.0
                dv1 = dv1 - dv_mvp
                dv2 = dv2 + dv_mvp

        # Primary Layers (Cruising aircraft has priority and clmibing/descending solves. All conflicts solved horizontally)
        elif self.priocode == 'LAY1':
            dv_mvp[2] = 0.0
            # If aircraft 1 is cruising, and aircraft 2 is climbing/descending -> aircraft 2 solves conflict horizontally
            if abs(vs1) < 0.1 and abs(vs2) > 0.1:
                dv2 = dv2 + dv_mvp
            # If aircraft 2 is cruising, and aircraft 1 is climbing -> aircraft 1 solves conflict horizontally
            elif abs(vs2) < 0.1 and abs(vs1) > 0.1:
                dv1 = dv1 - dv_mvp
            else:  # both are climbing/descending/cruising -> both aircraft solves the conflict horizontally
                dv1 = dv1 - dv_mvp
                dv2 = dv2 + dv_mvp

        # Secondary Layers (Climbing/descending aircraft has priority and cruising solves. All conflicts solved horizontally)
        elif self.priocode == 'LAY2':
            dv_mvp[2] = 0.0
            # If aircraft 1 is cruising, and aircraft 2 is climbing/descending -> aircraft 1 solves conflict horizontally
            if abs(vs1) < 0.1 and abs(vs2) > 0.1:
                dv1 = dv1 - dv_mvp
            # If aircraft 2 is cruising, and aircraft 1 is climbing -> aircraft 2 solves conflict horizontally
            elif abs(vs2) < 0.1 and abs(vs1) > 0.1:
                dv2 = dv2 + dv_mvp
            else:  # both are climbing/descending/cruising -> both aircraft solves the conflic horizontally
                dv1 = dv1 - dv_mvp
                dv2 = dv2 + dv_mvp

        return dv1, dv2


    def resolve(self, conf, ownship, intruder):
        ''' Resolve all current conflicts '''
        # Initialize an array to store the resolution velocity vector for all A/C
        dv = np.zeros((ownship.ntraf, 3))

        # Initialize an array to store time needed to resolve vertically
        timesolveV = np.ones(ownship.ntraf) * 1e9

        # Call MVP function to resolve conflicts-----------------------------------
        for ((ac1, ac2), qdr, dist, tcpa, tLOS) in zip(conf.confpairs, conf.qdr, conf.dist, conf.tcpa, conf.tLOS):
            idx1 = ownship.id.index(ac1)
            idx2 = intruder.id.index(ac2)

            # If A/C indexes are found, then apply MVP on this conflict pair
            # Because ADSB is ON, this is done for each aircraft separately
            if idx1 >-1 and idx2 > -1:
                dv_mvp, tsolV = self.MVP(ownship, intruder, conf, qdr, dist, tcpa, tLOS, idx1, idx2)
                if tsolV < timesolveV[idx1]:
                    timesolveV[idx1] = tsolV

                # Use priority rules if activated
                if self.swprio:
                    dv[idx1], _ = self.applyprio(dv_mvp, dv[idx1], dv[idx2], ownship.vs[idx1], intruder.vs[idx2])
                else:
                    # since cooperative, the vertical resolution component can be halved, and then dv_mvp can be added
                    dv_mvp[2] = 0.5 * dv_mvp[2]
                    dv[idx1] = dv[idx1] - dv_mvp

                # Check the noreso aircraft. Nobody avoids noreso aircraft.
                # But noreso aircraft will avoid other aircraft
                if self.noresoac[idx2]:
                    dv[idx1] = dv[idx1] + dv_mvp

                # Check the resooff aircraft. These aircraft will not do resolutions.
                if self.resooffac[idx1]:
                    dv[idx1] = 0.0


        # Determine new speed and limit resolution direction for all aicraft-------

        # Resolution vector for all aircraft, cartesian coordinates
        dv = np.transpose(dv)

        # The old speed vector, cartesian coordinates
        v = np.array([ownship.gseast, ownship.gsnorth, ownship.vs])

        # The new speed vector, cartesian coordinates
        newv = v + dv

        # Limit resolution direction if required-----------------------------------

        # Compute new speed vector in polar coordinates based on desired resolution
        if self.swresohoriz: # horizontal resolutions
            if self.swresospd and not self.swresohdg: # SPD only
                newtrack = ownship.trk
                newgs    = np.sqrt(newv[0,:]**2 + newv[1,:]**2)
                newvs    = ownship.vs
            elif self.swresohdg and not self.swresospd: # HDG only
                newtrack = (np.arctan2(newv[0,:],newv[1,:])*180/np.pi) % 360
                newgs    = ownship.gs
                newvs    = ownship.vs
            else: # SPD + HDG
                newtrack = (np.arctan2(newv[0,:],newv[1,:])*180/np.pi) %360
                newgs    = np.sqrt(newv[0,:]**2 + newv[1,:]**2)
                newvs    = ownship.vs
        elif self.swresovert: # vertical resolutions
            newtrack = ownship.trk
            newgs    = ownship.gs
            newvs    = newv[2,:]
        else: # horizontal + vertical
            newtrack = (np.arctan2(newv[0,:],newv[1,:])*180/np.pi) %360
            newgs    = np.sqrt(newv[0,:]**2 + newv[1,:]**2)
            newvs    = newv[2,:]

        # Determine ASAS module commands for all aircraft--------------------------

        # Cap the velocity
        newgscapped = np.maximum(ownship.perf.vmin,np.minimum(ownship.perf.vmax,newgs))

        # Cap the vertical speed
        vscapped = np.maximum(ownship.perf.vsmin,np.minimum(ownship.perf.vsmax,newvs))

        # Calculate if Autopilot selected altitude should be followed. This avoids ASAS from
        # climbing or descending longer than it needs to if the autopilot leveloff
        # altitude also resolves the conflict. Because asasalttemp is calculated using
        # the time to resolve, it may result in climbing or descending more than the selected
        # altitude.
        asasalttemp = vscapped * timesolveV + ownship.alt
        signdvs = np.sign(vscapped - ownship.ap.vs * np.sign(ownship.selalt - ownship.alt))
        signalt = np.sign(asasalttemp - ownship.selalt)
        alt = np.where(np.logical_or(signdvs == 0, signdvs == signalt), asasalttemp, ownship.selalt)

        # To compute asas alt, timesolveV is used. timesolveV is a really big value (1e9)
        # when there is no conflict. Therefore asas alt is only updated when its
        # value is less than the look-ahead time, because for those aircraft are in conflict
        altCondition = np.logical_and(timesolveV<conf.dtlookahead, np.abs(dv[2,:])>0.0)
        alt[altCondition] = asasalttemp[altCondition]

        # If resolutions are limited in the horizontal direction, then asasalt should
        # be equal to auto pilot alt (aalt). This is to prevent a new asasalt being computed
        # using the auto pilot vertical speed (ownship.avs) using the code in line 106 (asasalttemp) when only
        # horizontal resolutions are allowed.
        alt = alt * (1 - self.swresohoriz) + ownship.selalt * self.swresohoriz
        return newtrack, newgscapped, vscapped, alt

    def MVP(self, ownship, intruder, conf, qdr, dist, tcpa, tLOS, idx1, idx2):
        """Modified Voltage Potential (MVP) resolution method"""
        # Preliminary calculations-------------------------------------------------
        # Determine largest RPZ and HPZ of the conflict pair, use lookahead of ownship
        rpz_m = np.max(conf.rpz[[idx1, idx2]] * self.resofach)
        hpz_m = np.max(conf.hpz[[idx1, idx2]] * self.resofacv)
        dtlook = conf.dtlookahead[idx1]
        # Convert qdr from degrees to radians
        qdr = np.radians(qdr)

        # Relative position vector between id1 and id2
        drel = np.array([np.sin(qdr) * dist, \
                        np.cos(qdr) * dist, \
                        intruder.alt[idx2] - ownship.alt[idx1]])

        # Write velocities as vectors and find relative velocity vector
        v1 = np.array([ownship.gseast[idx1], ownship.gsnorth[idx1], ownship.vs[idx1]])
        v2 = np.array([intruder.gseast[idx2], intruder.gsnorth[idx2], intruder.vs[idx2]])
        vrel = v2 - v1


        # Horizontal resolution----------------------------------------------------

        # Find horizontal distance at the tcpa (min horizontal distance)
        dcpa  = drel + vrel*tcpa
        dabsH = np.sqrt(dcpa[0] * dcpa[0] + dcpa[1] * dcpa[1])

        # Compute horizontal intrusion
        iH = rpz_m - dabsH

        # Exception handlers for head-on conflicts
        # This is done to prevent division by zero in the next step
        if dabsH <= 10.:
            dabsH = 10.
            dcpa[0] = drel[1] / dist * dabsH
            dcpa[1] = -drel[0] / dist * dabsH

        # If intruder is outside the ownship PZ, then apply extra factor
        # to make sure that resolution does not graze IPZ
        if rpz_m < dist and dabsH < dist:
            # Compute the resolution velocity vector in horizontal direction.
            # abs(tcpa) because it bcomes negative during intrusion.
            erratum = np.cos(np.arcsin(rpz_m / dist)-np.arcsin(dabsH / dist))
            dv1 = ((rpz_m / erratum - dabsH) * dcpa[0]) / (abs(tcpa) * dabsH)
            dv2 = ((rpz_m / erratum - dabsH) * dcpa[1]) / (abs(tcpa) * dabsH)
        else:
            dv1 = (iH * dcpa[0]) / (abs(tcpa) * dabsH)
            dv2 = (iH * dcpa[1]) / (abs(tcpa) * dabsH)

        # Vertical resolution------------------------------------------------------

        # Compute the  vertical intrusion
        # Amount of vertical intrusion dependent on vertical relative velocity
        iV = hpz_m if abs(vrel[2]) > 0.0 else hpz_m - abs(drel[2])

        # Get the time to solve the conflict vertically - tsolveV
        tsolV = abs(drel[2] / vrel[2]) if abs(vrel[2]) > 0.0 else tLOS

        # If the time to solve the conflict vertically is longer than the look-ahead time,
        # because the the relative vertical speed is very small, then solve the intrusion
        # within tinconf
        if tsolV > dtlook:
            tsolV = tLOS
            iV    = hpz_m

        # Compute the resolution velocity vector in the vertical direction
        # The direction of the vertical resolution is such that the aircraft with
        # higher climb/decent rate reduces their climb/decent rate
        dv3 = np.where(abs(vrel[2]) > 0.0, (iV / tsolV) * (-vrel[2] / abs(vrel[2])), (iV / tsolV))

        # It is necessary to cap dv3 to prevent that a vertical conflict
        # is solved in 1 timestep, leading to a vertical separation that is too
        # high (high vs assumed in traf). If vertical dynamics are included to
        # aircraft  model in traffic.py, the below three lines should be deleted.
        #    mindv3 = -400*fpm# ~ 2.016 [m/s]
        #    maxdv3 = 400*fpm
        #    dv3 = np.maximum(mindv3,np.minimum(maxdv3,dv3))


        # Combine resolutions------------------------------------------------------

        # combine the dv components
        dv = np.array([dv1, dv2, dv3])

        return dv, tsolV



import pygame as pg

class Fastfont:
    """ 
    Faster font for printing graphically, renders only at creation

    Methods:
        CFastfont(screen,name,size,color,bold,italic) :  constructor, renders font
        
        printat(screen,x,y,text): print text at x,y, at screen window (blit)
        setpos(x,y)   : not used anymore?

    Members: see create

    Created by  : Jacco M. Hoekstra
    """

    def __init__(self,screen,name,size,color,bold,italic):
        self.swposx = -1  # Default x = left side
        self.swposy = -1  # Default y = top

        pfont = pg.font.SysFont(name,size,bold,italic)

        # Render font

        # Convert chars 32-126 to map of standard height bitmaps:
        self.chmaps = []
        self.chw = []
        self.linedy = pfont.get_linesize()
        
        for ich in range(32,126):
            ch = pfont.render(chr(ich),False,color)
            ch = ch.convert_alpha(screen)
            self.chmaps.append(ch)
            self.chw.append(ch.get_width())
        del pfont
        return      

    def printat(self,screen,x,y,text):

        w = 0
        for ch in text:
            ich = ord(ch)
            if ich>=32 and ich<=126:
                w = w + self.chw[ich-32]

        txtimg = pg.Surface((w,self.linedy)) # Standard height bitmap
        txtimg = txtimg.convert_alpha(screen)
        
        ix = 0
        for ch in text:
            ich = ord(ch)
            if ich>=32 and ich<=126:
                w = w + self.chw[ich-32]
                dest = self.chmaps[ich-32].get_rect()
                dest.top = 0
                dest.left = ix
                ix = ix+self.chw[ich-32]
                # chimg = self.chmaps[ich-32].convert_alpha(screen)

                txtimg.blit(self.chmaps[ich-32],dest)# Removed pg.BLEND_ADD which broke it on Windows machine

        dest = txtimg.get_rect()

        # Set position
        if self.swposx <0:
            dest.left = x
        elif self.swposx>0:
            dest.right = x
        else:
            dest.centerx = x

        if self.swposy <0:
            dest.top = y
        elif self.swposy>0:
            dest.bottom = y
        else:
            dest.centery = y

        # Paste it onto the screen
        screen.blit(txtimg,dest)# Removed pg.BLEND_ADD which broke it on Windows machine

        return

    def setpos(self,swposx,swposy):
        self.swposx = swposx
        self.swposy = swposy
        return



''' BlueSky shared state classes and functions.

    BlueSky's sharedstate is used to keep a shared state 
    across client(s) and simulation node(s)
'''
from typing import Any, Generic, Optional, Type, TypeVar
import numpy as np

from types import SimpleNamespace
from copy import deepcopy
from collections import defaultdict

import bluesky as bs
from bluesky.core import signal
from bluesky.network import context as ctx
from bluesky.network.common import ActionType


# Keep track of the set of subscribed sharedstate topics. Store signals to emit
# whenever a state update of each topic is received
sigchanged: dict[str, signal.Signal] = dict()


def reset(remote_id=None):
    ''' Reset shared state data to defaults for remote simulation. '''
    remotes[remote_id or bs.net.act_id] = _genstore()

    # If this is the active node, also emit a signal about this change
    if ctx.sender_id == bs.net.act_id:
        ctx.action = ActionType.Reset
        ctx.action_content = None
        for topic, sig in sigchanged.items():
            store = get(group=topic.lower())
            sig.emit(store)
        ctx.action = ActionType.NoAction


@signal.subscriber(topic='actnode-changed')
def on_actnode_changed(act_id):
    ctx.action = ActionType.ActChange
    ctx.action_content = None
    for topic, sig in sigchanged.items():
            store = get(group=topic.lower())
            sig.emit(store)
    ctx.action = ActionType.NoAction


def on_sharedstate_received(action, data):
    ''' Retrieve and process state data. '''
    store = get(ctx.sender_id, ctx.topic.lower())

    # Store sharedstate context
    ctx.action = ActionType(action)
    ctx.action_content = data

    if ctx.action == ActionType.Update:
        store.update(data)

    elif ctx.action == ActionType.Append:
        store.append(data)

    elif ctx.action == ActionType.Extend:
        store.extend(data)

    elif ctx.action == ActionType.Replace:
        store.replace(data)

    elif ctx.action == ActionType.Delete:
        store.delete(data)

    # Inform subscribers of state update
    # TODO: what to do with act vs all?
    if ctx.sender_id == bs.net.act_id:
        sigchanged[ctx.topic.lower()].emit(store)

    # Reset context variables
    ctx.action = ActionType.NoAction
    ctx.action_content = None


def get(remote_id=None, group=None):
    ''' Retrieve a remote store, or a group in a remote store.
        Returns the store of the active remote if no remote id is provided.
    '''
    return (remotes[remote_id or bs.net.act_id] if group is None else
            getattr(remotes[remote_id or bs.net.act_id], group))


def setvalue(name, value, remote_id=None, group=None):
    ''' Set the value of attribute 'name' in group 'group' for remote store with id 'remote_id' 
        Sets value in store of the active remote if no remote_id is provided.
    '''
    setattr(remotes[remote_id or bs.net.act_id] if group is None else
            getattr(remotes[remote_id or bs.net.act_id], group), name, value)


def setdefault(name: str, default: Any, group: Optional[str]=None):
    ''' Set the default value for variable 'name' in group 'group' '''
    # Add group if it doesn't exist yet
    if group is not None:
        addtopic(group)

    target = getattr(defaults, group) if group else defaults
    if not hasattr(target, name):
        # In case remote data already exists and this is a previously unknown variable, update stores
        for remote_id in remotes.keys():
            setvalue(name, deepcopy(default), remote_id, group)
    setattr(target, name, default)


def addtopic(topic: str) -> signal.Signal:
    ''' Add a sharedstate topic if it doesn't yet exist.
    
        This creates a storage group for this topic, which is added to each 
        remote data store.
        
        Arguments:
        - topic:  The sharedstate topic to add
    '''
    topic = topic.lower()

    # Create/get the signal that is emitted when this data changes
    sig = signal.Signal(f'state-changed.{topic}')
    sigchanged[topic] = sig

    # No creation needed if topic is already known
    if not hasattr(defaults, topic):
        # Add store to the defaults
        setattr(defaults, topic, Store())

        # Also add to existing stores if necessary
        for remote in remotes.values():
            setattr(remote, topic, Store())

    return sig


def is_sharedstate(topic):
    ''' Inform whether 'topic' is (already) known as sharedstate. '''
    return topic.lower() in sigchanged


class Store(SimpleNamespace):
    ''' Simple storage object for nested storage of state data per simulation node. '''
    def valid(self):
        ''' Return True if this store has initialised attributes.'''
        return all([bool(v) for v in vars(self).values()] or [False])

    def update(self, data):
        ''' Update a value in this store. '''
        for key, item in data.items():
            container = getattr(self, key, None)
            if container is None:
                setattr(self, key, item)
            elif isinstance(container, dict):
                _recursive_update(container, item)
            else:
                for idx, value in item.items():
                    container[idx] = value

    def append(self, data):
        ''' Append data to (lists/arrays in) this store. '''
        for key, item in data.items():
            container = getattr(self, key, None)
            if container is None:
                setattr(self, key, [item])
            elif isinstance(container, np.ndarray):
                setattr(self, key, np.append(container, item))

    def extend(self, data):
        ''' Extend data in (lists/arrays in) this store. '''
        for key, item in data.items():
            container = getattr(self, key, None)
            if container is None:
                setattr(self, key, item)
            elif isinstance(container, list):
                container.extend(item)
            elif isinstance(container, np.ndarray):
                setattr(self, key, np.concatenate([container, item]))

    def replace(self, data):
        ''' Replace data containers in this store. '''
        vars(self).update(data)

    def delete(self, data):
        ''' Delete data from this store. '''
        # We are expecting either an index, or a key value from a reference variable
        for key, item in data.items():
            idx = None
            if key not in vars(self):
                # Assume we are receiving an index to arrays/lists in this store
                if not isinstance(item, int):
                    raise ValueError(f"Expected integer index for delete {key} in topic {ctx.topic}") 
                idx = item
            else:
                ref = getattr(self, key)
                # If ref is a dict, this delete action should only act on the dict.
                if isinstance(ref, dict):
                    if isinstance(item, (list, tuple)):
                        for key in item:
                            ref.pop(key, None)
                    else:
                        ref.pop(item, None)
                    continue
                # Otherwise, assume we are receiving a lookup key for a reference value
                elif isinstance(ref, np.ndarray):
                    indices = np.where(ref == item)[0]
                    if not indices:
                        raise KeyError(f'Item with key {item} not found for variable {key} in topic {ctx.topic}')
                    idx = indices[0] if len(indices) == 1 else indices
                elif isinstance(ref, list):
                    idx = ref.index(item)

            if idx is None:
                continue

            for container in vars(self).values():
                if isinstance(container, np.ndarray):
                    mask = np.ones_like(container, dtype=bool)
                    mask[idx] = False
                    setattr(self, key, container[mask])
                elif isinstance(container, list):
                    if isinstance(idx, int):
                        container.pop(idx)
                    else:
                        # Assume a tuple or list
                        for iidx in reversed(sorted(idx)):
                            container.pop(iidx)


# Type template variable for ActData accessor class
T = TypeVar('T')

class ActData(Generic[T]):
    ''' Access shared state data from the active remote as if it is a member variable. '''
    __slots__ = ('default', 'name', 'group')

    def __init__(self, default=None, name='', group=None):
        self.default = default
        self.name: str = name
        self.group: Optional[str] = group

    def __class_getitem__(cls, key):
        ''' Make sure that the final annotation is equal to the wrapped type '''
        return key

    def __set_name__(self, owner, name):
        if not self.name:
            # Get name from attribute name if not previously specified
            self.name = name

        # If we have a default value, and it is not centrally known yet
        # store it for Store generation
        store = defaults if self.group is None else getattr(defaults, self.group, None)
        if store is None or not hasattr(store, name):
            if self.default is not None:
                # If specified use our default
                setdefault(name, self.default, self.group)
            else:
                # Otherwise try to get the default type from annotation
                # Look for __origin__ in case we have a GenericAlias like list[int]
                tp = owner.__annotations__.get(name)
                tp = getattr(tp, '__origin__', tp)
                if isinstance(tp, Type):
                    # Exception case for NumPy ndarray,
                    # which has shape as a mandatory ctor argument
                    if tp is np.ndarray:
                        setdefault(name, np.ndarray(0), self.group)
                    else:
                        setdefault(name, tp(), self.group)

    def __get__(self, obj, objtype=None) -> T:
        ''' Return the actual value for the currently active node. '''
        store = remotes[bs.net.act_id] if bs.net.act_id else defaults
        return getattr(store if self.group is None else getattr(store, self.group), self.name)
        # raise KeyError(f'ActData: {self.name} not found, group={self.group}, active node id={bs.net.act_id}')
        # TODO: What is the active (client) node on the sim-side? Is this always within a currently processed stack command? -> stack.sender_id

    def __set__(self, obj, value: T):
        if not bs.net.act_id:
            self.default = value
        else:
            setattr(remotes[bs.net.act_id] if self.group is None else getattr(remotes[bs.net.act_id], self.group), self.name, value)


def _recursive_update(target, source):
    ''' Recursively update nested dicts/lists/arrays in a Store. '''
    for k, v in source.items():
        if isinstance(v, dict):
            inner = target.get(k)
            if inner is not None:
                _recursive_update(inner, v)
                continue
        target[k] = v


def _genstore():
    ''' Generate a store object for a remote simulation from defaults. '''
    store = deepcopy(defaults)
    return store


# Keep track of default attribute values of mutable type.
# These always need to be stored per remote node.
defaults = Store()

# Keep a dict of remote state storage namespaces
remotes = defaultdict(_genstore)



import numpy as np

def encode_ndarray(o):
    '''Msgpack encoder for numpy arrays.'''
    if isinstance(o, np.ndarray):
        return {b'numpy': True,
                b'type': o.dtype.str,
                b'shape': o.shape,
                b'data': o.tobytes()}
    return o

def decode_ndarray(o):
    '''Msgpack decoder for numpy arrays.'''
    if o.get(b'numpy'):
        return np.fromstring(o[b'data'], dtype=np.dtype(o[b'type'])).reshape(o[b'shape'])
    return o



""" BlueSky plugin template. The text you put here will be visible
    in BlueSky as the description of your plugin. """
import numpy as np
# Import the global bluesky objects. Uncomment the ones you need
from bluesky import stack, ui  #, settings, navdb, sim, scr, tools
from bluesky.ui.qtgl.glhelpers import gl, RenderObject, VertexArrayObject


### Initialization function of your plugin. Do not change the name of this
### function, as it is the way BlueSky recognises this file as a plugin.
def init_plugin():
    ''' Plugin initialisation function. '''
    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'GUIEXAMPLE',

        # The type of this plugin.
        'plugin_type':     'gui',
        }

    # init_plugin() should always return a configuration dict.
    return config


### Entities in BlueSky are objects that are created only once (called singleton)
### which implement some traffic or other simulation functionality.
### To define an entity that ADDS functionality to BlueSky, create a class that
### inherits from bluesky.core.Entity.
### To replace existing functionality in BlueSky, inherit from the class that
### provides the original implementation (see for example the asas/eby plugin).
class MyVisual(RenderObject, layer=100):
    ''' Example new render object for BlueSky. '''
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.shape = VertexArrayObject(gl.GL_TRIANGLE_FAN)

    def create(self):
        vertices = np.array([52, 5, 52, 4, 54, 4, 54, 5], dtype = np.float32)
        self.shape.create(vertex=vertices, color=(255, 0, 0))

    def draw(self):
        self.shaderset.set_vertex_scale_type(self.shaderset.VERTEX_IS_LATLON)
        self.shape.draw()

    # You can create new stack commands with the stack.command decorator.
    # By default, the stack command name is set to the function name.
    # The default argument type is a case-sensitive word. You can indicate different
    # types using argument annotations. This is done in the below function:
    # - The color argument returns three integer values (r, g, b), which is why the 
    #   starred notation is used here.
    @stack.command
    def mycolor(self, *colour: 'colour'):
        ''' Set the color of my example shape. '''
        self.shape.single_color[1:4] = colour



''' Reference values for simulation data. '''
from bluesky.core.base import Base
from bluesky import stack
from bluesky.tools import areafilter


class RefData(Base):
    def __init__(self) -> None:
        super().__init__()
        self.lat: float = 0.0
        self.lon: float = 0.0
        self.alt: float = 0.0
        self.acidx: int = -1
        self.hdg: float = 0.0
        self.cas: float = 0.0
        self.area: areafilter.Shape = areafilter.Box('refarea', (-1.0, -1.0, 1.0, 1.0))

    def reset(self):
        ''' Reset reference data. '''
        self.lat = 0.0
        self.lon = 0.0
        self.alt = 0.0
        self.acidx = -1
        self.hdg = 0.0
        self.cas = 0.0
        self.area = areafilter.Box('refarea', (-1.0, -1.0, 1.0, 1.0))

    @stack.command
    def near(self, lat: 'lat', lon: 'lon', cmdstr: 'string'):
        '''Set reference lat/lon before executing command string. '''
        self.lat = lat
        self.lon = lon
        stack.process(cmdstr)

    @stack.commandgroup
    def inside(self, lat0: 'lat', lon0: 'lon', lat1: 'lat', lon1: 'lon', cmdstr: 'string'):
        self.area = areafilter.Box('refarea', (lat0, lon0, lat1, lon1))
        stack.process(cmdstr)

    @inside.altcommand
    def insidearea(self, areaname: 'txt', cmdstr: 'string'):
        self.area = areafilter.getArea(areaname)
        stack.process(cmdstr)

    @stack.command(aliases=('with',))
    def withac(self, idx: 'acid', cmdstr: 'string'):
        self.acidx = idx
        stack.process(cmdstr)



""" Documentation window for the QTGL version of BlueSky."""
from PyQt6.QtCore import QUrl, QFileInfo
from PyQt6.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QPushButton, QLabel
try:
    from PyQt6.QtWebEngineWidgets import QWebEngineView as QWebView
    from PyQt6.QtWebEngineCore import QWebEnginePage as QWebPage


    class DocView(QWebView):
        def __init__(self, parent=None):
            super().__init__(parent)
            class DocPage(QWebPage):
                def acceptNavigationRequest(self, url, navtype, ismainframe):
                    if navtype == self.NavigationType.NavigationTypeLinkClicked:
                        if url.url()[:6].lower() == 'stack:':
                            stack(url.url()[6:].lower())
                            return False
                    return True
            self.docpage = DocPage()
            self.setPage(self.docpage)
except ImportError:
    DocView = None

import bluesky as bs
from bluesky.stack import stack


class DocWindow(QWidget):
    app = None

    def __init__(self, app):
        super().__init__()
        self.vlayout  = QVBoxLayout()
        self.backbtn = QPushButton('Back')
        self.closebtn = QPushButton('Close')
        if DocView is not None:
            self.view = DocView()
            self.backbtn.clicked.connect(self.view.back)
        else:
            self.view = QLabel('BlueSky was not able to initialize it\'s\n' +
                ' QtWebEngine-based documentation viewer.\n' +
                'There may be something wrong with your Qt installation.\n' +
                'If you haven\'t yet, try installing PyQtWebEngine:\n\n' +
                '    pip install PyQtWebEngine\n\n' +
                'or, if you don\'t use pip, install it with your preferred\n' +
                'python package manager.')
        self.vlayout.setContentsMargins(1, 1, 1, 1)
        self.vlayout.setSpacing(1)
        self.vlayout.addWidget(self.view)
        hlayout = QHBoxLayout()
        buttonbox = QWidget()
        buttonbox.setLayout(hlayout)
        self.vlayout.addWidget(buttonbox)
        hlayout.addWidget(self.closebtn)
        hlayout.addWidget(self.backbtn)
        self.closebtn.clicked.connect(self.hide)
        self.setLayout(self.vlayout)
        self.setWindowTitle('BlueSky documentation')

    def show_cmd_doc(self, cmd):
        if not cmd:
            cmd = 'Command-Reference'
        if not isinstance(self.view, QLabel):
            fname = bs.resource(f'html/{cmd.lower()}.html').as_posix()
            self.view.load(QUrl.fromLocalFile(QFileInfo(fname).absoluteFilePath()))



''' The BlueSky command stack.

    The stack parses all text-based commands in the simulation.
'''
from bluesky.stack.stackbase import stack, forward, sender, get_scenname, get_scendata, set_scendata
from bluesky.stack.cmdparser import command, commandgroup, append_commands, \
    remove_commands, get_commands
from bluesky.stack.argparser import ArgumentError


def process(ext_cmds):
    ''' Dummy process that will get replaced by sim or client implementation during init. '''
    pass


def echo(self, text='', flags=0):
    ''' Echo

        This base implementation just prints to the terminal, and is used in sim-detached mode. It is replaced
        by sim or client implementation during init.
    '''
    print(text)


def init(mode='client'):
    if mode[:3] == 'sim':
        import bluesky.stack.simstack as simstack
        from bluesky.stack.importer import Importer
        simstack.init()
        globals()['process'] = simstack.process
        if 'detached' not in mode:
            globals()['echo'] = simstack.echo
    else:
        import bluesky.stack.clientstack as clientstack
        clientstack.init()
        globals()['process'] = clientstack.process
        globals()['echo'] = clientstack.echo



""" Main window for the QTGL gui."""
from pathlib import Path
import platform

from PyQt6.QtWidgets import QApplication as app, QWidget, QMainWindow, \
    QSplashScreen, QTreeWidgetItem, QPushButton, QFileDialog, QDialog, \
    QTreeWidget, QVBoxLayout, QDialogButtonBox, QMenu, QLabel
from PyQt6.QtCore import Qt, pyqtSlot, QTimer, QItemSelectionModel, QSize, QEvent, pyqtProperty
from PyQt6.QtGui import QPixmap, QIcon
from PyQt6 import uic

# Local imports
import bluesky as bs
from bluesky.core import Base, Signal
from bluesky.network.discovery import Discovery

from bluesky import stack
from bluesky.stack.argparser import PosArg
from bluesky.pathfinder import ResourcePath
from bluesky.tools.misc import tim2txt
from bluesky.network import subscriber, context as ctx
from bluesky.network.common import get_ownip, seqidx2id, seqid2idx
import bluesky.network.sharedstate as ss

from bluesky.ui import palette

# Child windows
from bluesky.ui.qtgl.docwindow import DocWindow
from bluesky.ui.qtgl.infowindow import InfoWindow
from bluesky.ui.qtgl.settingswindow import SettingsWindow
# from bluesky.ui.qtgl.nd import ND

if platform.system().lower() == "windows":
    from bluesky.ui.pygame.dialog import fileopen
    import winreg
    def isdark():
        ''' Returns true if app is in dark mode, false otherwise. '''
        try:
            registry = winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)
            key = winreg.OpenKey(registry, r"Software\\Microsoft\\\Windows\\CurrentVersion\\Themes\\Personalize")
            value, _ = winreg.QueryValueEx(key, "AppsUseLightTheme")
            return value == 0
        except FileNotFoundError:
            return False
else:
    def isdark():
        ''' Returns true if app is in dark mode, false otherwise. '''
        p = app.instance().style().standardPalette()
        return (p.color(p.ColorRole.Window).value() < p.color(p.ColorRole.WindowText).value())


# Register settings defaults
bs.settings.set_variable_defaults(gfx_path='graphics', start_location='EHAM')

palette.set_default_colours(stack_text=(0, 255, 0),
                            stack_background=(102, 102, 102))


class Splash(QSplashScreen):
    """ Splash screen: BlueSky logo during start-up"""
    def __init__(self):
        splashfile = bs.resource(bs.settings.gfx_path) / 'splash.gif'
        super().__init__(QPixmap(splashfile.as_posix()), Qt.WindowType.WindowStaysOnTopHint)


class DiscoveryDialog(QDialog):
    def __init__(self, comm_id, parent=None):
        super().__init__(parent)
        self.setModal(True)
        self.setMinimumSize(200,200) # To prevent Geometry error
        self.servers = []
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.serverview = QTreeWidget()
        self.serverview.setHeaderLabels(['Server', 'Ports'])
        self.serverview.setIndentation(0)
        self.serverview.setStyleSheet('padding:0px')
        self.serverview.header().resizeSection(0, 180)
        layout.addWidget(self.serverview)
        btns = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        layout.addWidget(btns)
        btns.accepted.connect(self.on_accept)
        btns.rejected.connect(parent.closeEvent)

        self.discovery = Discovery(comm_id)

        self.discovery_timer = QTimer()
        self.discovery_timer.timeout.connect(self.discovery.update)
        self.discovery_timer.start(1000)
        self.discovery.server_discovered.connect(self.add_srv)
        self.discovery.start()

    def add_srv(self, address, ports):
        for server in self.servers:
            if address == server.address and ports == server.ports:
                # We already know this server, skip
                return
        server = QTreeWidgetItem(self.serverview)
        server.address = address
        server.ports = ports
        server.hostname = 'This computer' if address == get_ownip() else address
        server.setText(0, server.hostname)

        server.setText(1, '{},{}'.format(*ports))
        self.servers.append(server)

    def on_accept(self):
        server = self.serverview.currentItem()
        if server:
            self.discovery_timer.stop()
            self.discovery.stop()
            hostname = server.address
            rport, sport = server.ports
            bs.net.connect(hostname=hostname, recv_port=rport, send_port=sport)
            self.close()


class MainWindow(QMainWindow, Base):
    """ Qt window process: from .ui file read UI window-definition of main window """

    modes = ['Init', 'Hold', 'Operate', 'End']

    # Per remote node attributes
    nconf_cur: ss.ActData[int] = ss.ActData(0, group='acdata')
    nconf_tot: ss.ActData[int] = ss.ActData(0, group='acdata')
    nlos_cur: ss.ActData[int] = ss.ActData(0, group='acdata')
    nlos_tot: ss.ActData[int] = ss.ActData(0, group='acdata')

    show_map: ss.ActData[bool] = ss.ActData(True)
    show_coast: ss.ActData[bool] = ss.ActData(True)
    show_wpt: ss.ActData[int] = ss.ActData(1)
    show_apt: ss.ActData[int] = ss.ActData(1)
    show_pz: ss.ActData[bool] = ss.ActData(False)
    show_traf: ss.ActData[bool] = ss.ActData(True)
    show_lbl: ss.ActData[int] = ss.ActData(2)

    @pyqtProperty(str)
    def style(self):
        ''' Returns "dark"" if app is in dark mode, "light" otherwise. '''
        return "dark" if self.darkmode else "light"

    def __init__(self, mode):
        super().__init__()
        # Running mode of this gui. Options:
        #  - server-gui: Normal mode, starts bluesky server together with gui
        #  - client: starts only gui in client mode, can connect to existing
        #    server.
        self.mode = mode
        self.running = True

        # self.nd = ND(shareWidget=self.radarwidget)
        self.infowin = InfoWindow()
        self.settingswin = SettingsWindow()
        self.darkmode = isdark()

        try:
            self.docwin = DocWindow(self)
        except Exception as e:
            print('Couldnt make docwindow:', e)
        # self.aman = AMANDisplay()
        

        gfxpath = bs.resource(bs.settings.gfx_path)

        if platform.system() == 'Darwin':
            app.instance().setWindowIcon(QIcon((gfxpath / 'bluesky.icns').as_posix()))
        else:
            app.instance().setWindowIcon(QIcon((gfxpath / 'icon.gif').as_posix()))

        uic.loadUi((gfxpath / 'mainwindow.ui').as_posix(), self)
        gltimer = QTimer(self)
        gltimer.timeout.connect(self.radarwidget.update)
        # gltimer.timeout.connect(self.nd.updateGL)
        gltimer.start(50)

        # If multiple scenario paths exist, add 'Open From' menu
        scenresource = bs.resource('scenario')
        if isinstance(scenresource, ResourcePath) and scenresource.nbases > 1:
            openfrom = QMenu('Open From', self.menuFile)
            self.menuFile.insertMenu(self.action_Save, openfrom)

            openpkg = openfrom.addAction('Package')
            openpkg.triggered.connect(lambda: self.show_file_dialog(scenresource.base(-1)))
            openusr = openfrom.addAction('User')
            openusr.triggered.connect(lambda: self.show_file_dialog(scenresource.base(0)))

        # Link menubar buttons
        self.action_Open.triggered.connect(self.show_file_dialog)
        self.action_Save.triggered.connect(self.buttonClicked)
        self.actionBlueSky_help.triggered.connect(self.show_doc_window)
        self.actionSettings.triggered.connect(self.settingswin.show)

        # Connect to io client's nodelist changed signal
        bs.net.node_added.connect(self.nodesChanged)
        bs.net.server_added.connect(self.serversChanged)

        # Tell BlueSky that this is the screen object for this client
        bs.scr = self

        # Signals we want to emit
        self.panzoom_event = Signal('state-changed.panzoom')

        # Set position default from settings
        lat, lon, _ = PosArg().parse(bs.settings.start_location)
        ss.setdefault('pan', [lat, lon], group='panzoom')


        # self.nodetree.setVisible(False)
        self.nodetree.setIndentation(0)
        self.nodetree.setColumnCount(2)
        self.nodetree.setStyleSheet('padding:0px')
        self.nodetree.setAttribute(Qt.WidgetAttribute.WA_MacShowFocusRect, False)
        self.nodetree.header().resizeSection(0, 130)
        self.nodetree.itemClicked.connect(self.nodetreeClicked)
        self.maxservnum = 0
        self.servers = dict()
        self.nodes = dict()
        self.actnode = ''

        self.splitter.setSizes([1, 0])
        self.splitter_2.setSizes([1, 0])
        self.setStyleSheet()

        # Remove keyboard focus from all children of self.databox
        # (not possible to do in QDesigner)
        def recursiveNoFocus(w):
            for child in w.findChildren(QWidget):
                recursiveNoFocus(child)
            w.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        recursiveNoFocus(self.databox)

    def setStyleSheet(self, contents=''):
        if not contents:
            with open(bs.resource(bs.settings.gfx_path) / 'bluesky.qss') as style:
                super().setStyleSheet(style.read())

    @stack.command
    def swrad(self, switch: 'txt', arg: int|None = None):
        ''' Switch on/off elements and background of map/radar view 
        
            Usage:
                SWRAD GEO/GRID/APT/VOR/WPT/LABEL/ADSBCOVERAGE/TRAIL/POLY [dt]/[value]
        '''
        match switch:
            case 'GEO':
                self.show_coast = not self.show_coast
            case 'SAT':
                self.show_map = not self.show_map
            case 'APT':
                if arg is not None:
                    self.show_apt = min(2,max(0,arg))
                else:
                    self.show_apt = (self.show_apt + 1) % 3
            case 'WPT':
                if arg is not None:
                    self.show_wpt = min(2,max(0,arg))
                else:
                    self.show_wpt = (self.show_wpt + 1) % 3
            case 'LABEL':
                if arg is not None:
                    self.show_lbl = min(2,max(0,arg))
                else:
                    self.show_lbl = (self.show_lbl + 1) % 3
            case 'SYM':
                if arg is None:
                    arg = 0 if self.show_pz else (2 if self.show_traf else 1)
                self.show_traf = arg > 0
                self.show_pz = arg > 1

    @stack.command
    def mcre(self, args: 'string'):
        """ Create one or more random aircraft in a specified area 
        
            When called from the client (gui), MCRE will use the current screen bounds as area to create aircraft in.
            When called from a scenario, the simulation reference area will be used.
        """
        if not args:
            return stack.forward('MCRE')

        stack.forward(f'INSIDE {" ".join(str(el) for el in bs.ref.area.bbox)} MCRE {args}')

    @stack.command(annotations='pandir/latlon', brief='PAN latlon/acid/airport/waypoint/LEFT/RIGHT/UP/DOWN')
    def pan(self, *args):
        "Pan screen (move view) to a waypoint, direction or aircraft"
        store = ss.get(group='panzoom')
        store.pan = list(args)
        self.panzoom_event.emit(store)
        return True

    @stack.commandgroup(annotations='float/txt', brief='ZOOM IN/OUT/factor')
    def zoom(self, factor: float):
        ''' ZOOM: Zoom in and out in the radar view. 
        
            Arguments:
            - factor: IN/OUT to zoom in/out by a factor sqrt(2), or
                      'factor' to set zoom to specific value.
        '''
        store = ss.get(group='panzoom')
        store.zoom = factor
        self.panzoom_event.emit(store)
        return True

    @zoom.subcommand(name='IN')
    def zoomin(self, factor:float|None=None):
        ''' ZOOM IN: change zoom level up, relative to previous value '''
        store = ss.get(group='panzoom')
        store.zoom *= (factor or 1.4142135623730951)
        self.panzoom_event.emit(store)
        return True

    @zoom.subcommand(name='OUT')
    def zoomout(self, factor:float|None=None):
        ''' ZOOM OUT: change zoom level down, relative to previous value '''
        store = ss.get(group='panzoom')
        store.zoom *= (factor or 0.7071067811865475)
        self.panzoom_event.emit(store)
        return True

    def getviewctr(self):
        return self.radarwidget.pan

    def getviewbounds(self): # Return current viewing area in lat, lon
        return self.radarwidget.viewportlatlon()

    def keyPressEvent(self, event):
        if event.modifiers() & Qt.KeyboardModifier.ShiftModifier \
                and event.key() in [Qt.Key.Key_Up, Qt.Key.Key_Down, Qt.Key.Key_Left, Qt.Key.Key_Right]:
            dlat = 1.0 / (self.radarwidget.zoom * self.radarwidget.ar)
            dlon = 1.0 / (self.radarwidget.zoom * self.radarwidget.flat_earth)
            if event.key() == Qt.Key.Key_Up:
                self.radarwidget.setpanzoom(pan=(dlat, 0.0), absolute=False)
            elif event.key() == Qt.Key.Key_Down:
                self.radarwidget.setpanzoom(pan=(-dlat, 0.0), absolute=False)
            elif event.key() == Qt.Key.Key_Left:
                self.radarwidget.setpanzoom(pan=(0.0, -dlon), absolute=False)
            elif event.key() == Qt.Key.Key_Right:
                self.radarwidget.setpanzoom(pan=(0.0, dlon), absolute=False)

        elif event.key() == Qt.Key.Key_Escape:
            self.closeEvent()

        elif event.key() == Qt.Key.Key_F11:  # F11 = Toggle Full Screen mode
            if not self.isFullScreen():
                self.showFullScreen()
            else:
                self.showNormal()

        else:
            # All other events go to the BlueSky console
            self.console.keyPressEvent(event)
        event.accept()

    @stack.command(name='QUIT', annotations='', aliases=('CLOSE', 'END', 'EXIT', 'Q', 'STOP'))
    def closeEvent(self, event=None):
        if self.running:
            self.running = False
            # Send quit to server if we own it
            if self.mode != 'client':
                bs.net.send(b'QUIT', to_group=bs.server.server_id)
            app.instance().closeAllWindows()
            # return True

    @subscriber
    def echo(self, text, flags=None, sender_id=None):
        refnode = sender_id or ctx.sender_id or bs.net.act_id
        # Always update the store
        store = ss.get(refnode)
        store.echotext.append(text)
        store.echoflags.append(flags)
        # Directly echo if message corresponds to active node
        if refnode == bs.net.act_id:
            return self.console.echo(text, flags)

    def changeEvent(self, event: QEvent):
        # Detect dark/light mode switch
        if event.type() == event.Type.PaletteChange and self.darkmode != isdark():
            self.darkmode = isdark()
            self.setStyleSheet()

        return super().changeEvent(event)


    def actnodedataChanged(self, nodeid, nodedata, changed_elems):
        if nodeid != self.actnode:
            self.actnode = nodeid
            node = self.nodes[nodeid]
            self.nodelabel.setText(f'<b>Node</b> {node.serv_num}:{node.node_num}')
            self.nodetree.setCurrentItem(node, 0, QItemSelectionModel.SelectionFlag.ClearAndSelect)

    def serversChanged(self, server_id):
        server = self.servers.get(server_id)
        if not server:
            server = QTreeWidgetItem(self.nodetree)
            self.maxservnum += 1
            server.serv_num = self.maxservnum
            server.server_id = server_id
            hostname = 'Ungrouped' if server_id == b'0' else 'This computer'
            f = server.font(0)
            f.setBold(True)
            server.setExpanded(True)
            if server_id != b'0':
                btn = QPushButton(self.nodetree)
                btn.server_id = server_id
                btn.setText(hostname)
                btn.setFlat(True)
                btn.setStyleSheet('font-weight:bold')
                icon = bs.resource(bs.settings.gfx_path) / 'icons/addnode.svg'
                btn.setIcon(QIcon(icon.as_posix()))
                btn.setIconSize(QSize(40 if server_id == b'0' else 24, 16))
                btn.setLayoutDirection(Qt.LayoutDirection.RightToLeft)
                btn.setMaximumHeight(16)
                btn.clicked.connect(self.buttonClicked)
                self.nodetree.setItemWidget(server, 0, btn)

                # Move nodes from ungrouped if they belong to this server
                ungrouped: QTreeWidgetItem = self.servers.get(b'0')
                ucount = 0
                if ungrouped:
                    for node in ungrouped.takeChildren():
                        if node.node_id[:-1] + seqidx2id(0) == server_id:
                            server.addChild(node)
                        else:
                            ungrouped.addChild(node)
                            ucount += 1
                    if not ucount:
                        ungrouped.setHidden(True)
            else:
                self.nodetree.setItemWidget(server, 0, QLabel(hostname, parent=self.nodetree))
            self.servers[server_id] = server


    def nodesChanged(self, node_id):
        if node_id not in self.nodes:
            #print(node_id, 'added to list')
            server_id = node_id[:-1] + seqidx2id(0)
            if server_id not in bs.net.servers:
                server_id = b'0'
            if server_id not in self.servers:
                self.serversChanged(server_id)
            server = self.servers.get(server_id)
            server.setHidden(False)
            node_num = seqid2idx(node_id[-1])
            node = QTreeWidgetItem(server)
            node.setText(0, f'{server.serv_num}:{node_num} <init>')
            node.setText(1, '00:00:00')
            node.node_id  = node_id
            node.node_num = node_num
            node.serv_num = server.serv_num

            self.nodes[node_id] = node

    @subscriber(topic='SHOWDIALOG')
    def on_showdialog_received(self, dialog, args=''):
        ''' Processing of events from simulation nodes. '''
        # dialog = data.get('dialog')
        # args   = data.get('args')
        if dialog == 'OPENFILE':
            self.show_file_dialog()
        elif dialog == 'DOC':
            self.show_doc_window(args)

    @subscriber(topic='SIMINFO')
    def on_siminfo_received(self, speed, simdt, simt, simutc, ntraf, state, scenname):
        simt = tim2txt(simt)[:-3]
        self.setNodeInfo(ctx.sender_id, simt, scenname)
        if ctx.sender_id == bs.net.act_id:
            self.siminfoLabel.setText(u'<b>t:</b> %s, <b>\u0394t:</b> %.2f, <b>Speed:</b> %.1fx, <b>UTC:</b> %s, <b>Mode:</b> %s, <b>Aircraft:</b> %d, <b>Conflicts:</b> %d/%d, <b>LoS:</b> %d/%d'
                % (simt, simdt, speed, simutc, self.modes[state], ntraf, self.nconf_cur, self.nconf_tot, self.nlos_cur, self.nlos_tot))

    def setNodeInfo(self, connid, time, scenname):
        node = self.nodes.get(connid)
        if node:
            node.setText(0, f'{node.serv_num}:{node.node_num} <{scenname}>')
            node.setText(1, time)

    @pyqtSlot(QTreeWidgetItem, int)
    def nodetreeClicked(self, item, column):
        if item in self.servers.values():
            item.setSelected(False)
            item.child(0).setSelected(True)
            bs.net.actnode(item.child(0).node_id)
        else:
            bs.net.actnode(item.node_id)


    @pyqtSlot()
    def buttonClicked(self):
        if self.sender() == self.zoomin:
            self.radarwidget.setpanzoom(zoom=1.4142135623730951, absolute=False)
        elif self.sender() == self.zoomout:
            self.radarwidget.setpanzoom(zoom=0.70710678118654746, absolute=False)
        elif self.sender() == self.pandown:
            self.radarwidget.setpanzoom(pan=(-0.5,  0.0), absolute=False)
        elif self.sender() == self.panup:
            self.radarwidget.setpanzoom(pan=( 0.5,  0.0), absolute=False)
        elif self.sender() == self.panleft:
            self.radarwidget.setpanzoom(pan=( 0.0, -0.5), absolute=False)
        elif self.sender() == self.panright:
            self.radarwidget.setpanzoom(pan=( 0.0,  0.5), absolute=False)
        elif self.sender() == self.ic:
            self.show_file_dialog()
        elif self.sender() == self.sameic:
            stack.stack('IC IC')
        elif self.sender() == self.hold:
            stack.stack('HOLD')
        elif self.sender() == self.op:
            stack.stack('OP')
        elif self.sender() == self.fast:
            stack.stack('FF')
        elif self.sender() == self.fast10:
            stack.stack('FF 0:0:10')
        elif self.sender() == self.showac:
            stack.stack('SHOWTRAF')
        elif self.sender() == self.showpz:
            stack.stack('SHOWPZ')
        elif self.sender() == self.showapt:
            stack.stack('SHOWAPT')
        elif self.sender() == self.showwpt:
            stack.stack('SHOWWPT')
        elif self.sender() == self.showlabels:
            stack.stack('LABEL')
        elif self.sender() == self.showmap:
            stack.stack('SHOWMAP')
        elif self.sender() == self.action_Save:
            stack.stack('SAVEIC')
        elif hasattr(self.sender(), 'server_id'):
            bs.net.send(b'ADDNODES', 1, self.sender().server_id)

    def show_file_dialog(self, path=None):
        # Due to Qt5 bug in Windows, use temporarily Tkinter
        if platform.system().lower()=='windows':
            fname = fileopen()
        else:
            if path is None or isinstance(path, bool):
                path = bs.resource(bs.settings.scenario_path)

            if isinstance(path, ResourcePath):
                def getscenpath(resource):
                    # Find first path that contains scenario files
                    for p in resource.bases():
                        for f in p.glob('*.[Ss][Cc][Nn]'):
                            if f.name.lower() != 'ic.scn':
                                return p.as_posix()
                    return p.as_posix()
                scenpath = getscenpath(path)
            elif isinstance(path, Path):
                scenpath = path.as_posix()
            else:
                scenpath = path
            
            if platform.system().lower() == 'darwin':
                response = QFileDialog.getOpenFileName(self, 'Open file', scenpath, 'Scenario files (*.scn)')
            else:
                response = QFileDialog.getOpenFileName(self, 'Open file', scenpath, 'Scenario files (*.scn)', options=QFileDialog.Option.DontUseNativeDialog)
            fname = response[0] if isinstance(response, tuple) else response

        # Send IC command to stack with filename if selected, else do nothing
        if fname:
            bs.stack.stack('IC ' + str(fname))

    def show_doc_window(self, cmd=''):
        self.docwin.show_cmd_doc(cmd)
        self.docwin.show()
